__all__ = ()

import re

from .utils import istr

METHOD_ANY = '*'
METHOD_CONNECT = 'CONNECT'
METHOD_HEAD = 'HEAD'
METHOD_GET = 'GET'
METHOD_DELETE = 'DELETE'
METHOD_OPTIONS = 'OPTIONS'
METHOD_PATCH = 'PATCH'
METHOD_POST = 'POST'
METHOD_PUT = 'PUT'
METHOD_TRACE = 'TRACE'

METHOD_GET_ALL = {METHOD_GET, METHOD_HEAD, METHOD_OPTIONS}
METHOD_POST_ALL = {METHOD_PATCH, METHOD_POST, METHOD_PUT}

METHOD_ALL = {METHOD_CONNECT, METHOD_HEAD, METHOD_GET, METHOD_DELETE, METHOD_OPTIONS, METHOD_PATCH, METHOD_POST,
    METHOD_PUT, METHOD_TRACE}


ACCEPT = istr('Accept')
ACCEPT_CHARSET = istr('Accept-Charset')
ACCEPT_ENCODING = istr('Accept-Encoding')
ACCEPT_LANGUAGE = istr('Accept-Language')
ACCEPT_RANGES = istr('Accept-Ranges')
ACCESS_CONTROL_MAX_AGE = istr('Access-Control-Max-Age')
ACCESS_CONTROL_ALLOW_CREDENTIALS = istr('Access-Control-Allow-Credentials')
ACCESS_CONTROL_ALLOW_HEADERS = istr('Access-Control-Allow-Headers')
ACCESS_CONTROL_ALLOW_METHODS = istr('Access-Control-Allow-Methods')
ACCESS_CONTROL_ALLOW_ORIGIN = istr('Access-Control-Allow-Origin')
ACCESS_CONTROL_EXPOSE_HEADERS = istr('Access-Control-Expose-Headers')
ACCESS_CONTROL_REQUEST_HEADERS = istr('Access-Control-Request-Headers')
ACCESS_CONTROL_REQUEST_METHOD = istr('Access-Control-Request-Method')
AGE = istr('Age')
ALLOW = istr('Allow')
AUTHORIZATION = istr('Authorization')
CACHE_CONTROL = istr('Cache-Control')
CONNECTION = istr('Connection')
CONTENT_DISPOSITION = istr('Content-Disposition')
CONTENT_ENCODING = istr('Content-Encoding')
CONTENT_LANGUAGE = istr('Content-Language')
CONTENT_LENGTH = istr('Content-Length')
CONTENT_LOCATION = istr('Content-Location')
CONTENT_MD5 = istr('Content-MD5')
CONTENT_RANGE = istr('Content-Range')
CONTENT_TRANSFER_ENCODING = istr('Content-Transfer-Encoding')
CONTENT_TYPE = istr('Content-Type')
COOKIE = istr('Cookie')
DATE = istr('Date')
DESTINATION = istr('Destination')
DIGEST = istr('Digest')
ETAG = istr('Etag')
EXPECT = istr('Expect')
EXPIRES = istr('Expires')
FORWARDED = istr('Forwarded')
FROM = istr('From')
HOST = istr('Host')
IF_MATCH = istr('If-Match')
IF_MODIFIED_SINCE = istr('If-Modified-Since')
IF_NONE_MATCH = istr('If-None-Match')
IF_RANGE = istr('If-Range')
IF_UNMODIFIED_SINCE = istr('If-Unmodified-Since')
KEEP_ALIVE = istr('Keep-Alive')
LAST_EVENT_ID = istr('Last-Event-ID')
LAST_MODIFIED = istr('Last-Modified')
LINK = istr('Link')
LOCATION = istr('Location')
MAX_FORWARDS = istr('Max-Forwards')
ORIGIN = istr('Origin')
PRAGMA = istr('Pragma')
PROXY_AUTHENTICATE = istr('Proxy-Authenticate')
PROXY_AUTHORIZATION = istr('Proxy-Authorization')
RANGE = istr('Range')
REFERER = istr('Referer')
RETRY_AFTER = istr('Retry-After')
SEC_WEBSOCKET_ACCEPT = istr('Sec-WebSocket-Accept')
SEC_WEBSOCKET_VERSION = istr('Sec-WebSocket-Version')
SEC_WEBSOCKET_PROTOCOL = istr('Sec-WebSocket-Protocol')
SEC_WEBSOCKET_EXTENSIONS = istr('Sec-WebSocket-Extensions')
SEC_WEBSOCKET_KEY = istr('Sec-WebSocket-Key')
SEC_WEBSOCKET_KEY1 = istr('Sec-WebSocket-Key1')
SERVER = istr('Server')
SET_COOKIE = istr('Set-Cookie')
TE = istr('TE')
TRAILER = istr('Trailer')
TRANSFER_ENCODING = istr('Transfer-Encoding')
UPGRADE = istr('Upgrade')
WEBSOCKET = istr('WebSocket')
URI = istr('URI')
USER_AGENT = istr('User-Agent')
VARY = istr('Vary')
VIA = istr('Via')
WANT_DIGEST = istr('Want-Digest')
WARNING = istr('Warning')
WWW_AUTHENTICATE = istr('WWW-Authenticate')
X_FORWARDED_FOR = istr('X-Forwarded-For')
X_FORWARDED_HOST = istr('X-Forwarded-Host')
X_FORWARDED_PROTO = istr('X-Forwarded-Proto')

_TOKEN_RP = re.compile(r'[-!#$%&\'*+.^_`|~0-9a-zA-Z]+')
_LIST_START_RP = re.compile('[\t ,]*')
_SPACE_RP = re.compile('[\t ]*')
_PROTOCOL_RP = re.compile(r'[-!#$%&\'*+.^_`|~0-9a-zA-Z]+(?:/[-!#$%&\'*+.^_`|~0-9a-zA-Z]+)?')

del re
del istr

def build_extensions(available_extensions):
    """
    Builds websocket extensions header from the given extension values.
    
    Parameters
    ----------
    available_extensions : `list` of `Any`
        Each websocket extension should have the following `4` attributes / methods:
        - `name`, type `str`. The extension's name.
        - `request_params` : `list` of `tuple` (`str`, `str`). Additional header parameters of the extension.
        - `decode` : `callable`. Decoder method, what processes a received websocket frame. Should accept `2`
            parameters: The respective websocket ``Frame``, and the ˙max_size` as `int`, what decides the
            maximal size of a received frame. If it is passed, ``PayloadError`` is raised.
        - `encode` : `callable`. Encoder method, what processes the websocket frames to send. Should accept `1`
            parameter, the respective websocket ``Frame``.
    
    Returns
    -------
    header_value : `str`
    """
    main_parts = []
    sub_parts = []
    
    for available_extension in available_extensions:
        name = available_extension.name
        parameters = available_extension.request_params
        
        sub_parts.append(name)
        for key, value in parameters:
            if value is None:
                sub_parts.append(key)
            else:
                sub_parts.append(f'{key}={value}')
        
        main_parts.append('; '.join(sub_parts))
        sub_parts.clear()
    
    return ', '.join(main_parts)

def parse_extensions(header_value):
    """
    Parses extension header.
    
    Parameters
    ----------
    header_value : `str`
        Received extension header.

    Returns
    -------
    result : `list` of `tuple` (`str`, `list` of `tuple` (`str`, `str`))
        The parsed out extensions as `name` - `parameters` pairs. The `parameters` are in `list` storing
        `key` - `value` pairs.
    
    Raises
    ------
    ValueError
        Extension header value is incorrect.
    """
    result = []
    limit = len(header_value)
    index = 0
    
    check_start = True
    
    while True:
        # parse till 1st element
        matched = _LIST_START_RP.match(header_value, index)
        index = matched.end()
        
        # are we at the end?
        if index == limit:
            return result
        
        # now lets parse the extension's name
        matched = _TOKEN_RP.match(header_value, index)
        if matched is None:
            raise ValueError(f'Expected extension name since index {index}.')
        name = matched.group(0)
        index = matched.end()
        
        # nice, we have a name, we can make our item now!
        sub_parts = []
        result.append((name, sub_parts,),)
        
        # should we parse a sublist?
        while True:
            
            # after half item we skip this part
            if check_start:
                # are we at the end?
                if index == limit:
                    return result
                
                # lets parse till next character
                matched = _SPACE_RP.match(header_value, index)
                index = matched.end()
                
                # are we at the end?
                if index == limit:
                    return result
                
                # no sublist?
                if header_value[index] == ',':
                    index += 1
                    break

                # invalid character
                if header_value[index] != ';':
                    raise ValueError(f'Expected \';\' at index {index}.')
                
                # we have a sublist
                index += 1
                
            else:
                check_start = True
            
            # parse space
            matched = _SPACE_RP.match(header_value, index)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                break
            
            # lets parse the key now
            matched = _TOKEN_RP.match(header_value, index)
            if matched is None:
                raise ValueError(f'Expected parameter name since index {index}.')
            key = matched.group(0)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                sub_parts.append((key, None,),)
                break
            
            # parse space
            matched = _SPACE_RP.match(header_value, index)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                sub_parts.append((key, None,),)
                break

            #is it a full item or a half?
            
            #next extension
            if header_value[index] == ',':
                sub_parts.append((key, None,),)
                index += 1
                break

            # next item
            if header_value[index] == ';':
                sub_parts.append((key, None,),)
                index += 1
                check_start = False
                continue

            #invalid character
            if header_value[index] != '=':
                raise ValueError(f'Expected \',\' or \';\' or \'=\' at index {index}.')
            
            index += 1
            
            # parse space
            matched = _SPACE_RP.match(header_value, index)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                raise ValueError('Expected a parameter value, but string ended.')
            
            # is it '"stuff"' ?
            if header_value[index] == '"':
                index += 1
                
                # are we at the end?
                if index == limit:
                    raise ValueError('Expected a parameter value, but string ended.')
                
                matched = _TOKEN_RP.match(header_value, index)
                if matched is None:
                    raise ValueError(f'Expected parameter value since index {index}.')
                value = matched.group(0)
                index = matched.end()
                
                # are we at the end? or did we finish the string normally?
                if index == limit or header_value[index] != '"':
                    raise ValueError('Expected a \'"\' after starting a value with \'"\'.')
                index += 1
            
            # is it 'stuff' ?
            else:
                matched = _TOKEN_RP.match(header_value, index)
                if matched is None:
                    raise ValueError(f'Expected parameter value since index {index}.')
                value = matched.group(0)
                index = matched.end()
            
            # we got a full item
            sub_parts.append((key, value,),)

def parse_connections(header_value):
    """
    Parses subprotocol or connection headers.
    
    Parameters
    ----------
    header_value : `str`
        Received subprotocol or connection header.
    
    Returns
    -------
    result : `list` of `str`
        The parsed subprotocol or connection headers.
    
    Raises
    ------
    ValueError
        Subprotocol or connection header value is incorrect.
    """
    result = []
    limit = len(header_value)
    index = 0
    
    while True:
        #parse till 1st element
        matched = _LIST_START_RP.match(header_value, index)
        index = matched.end()
        
        #are we at the end?
        if index == limit:
            return result
        
        #now lets parse the upgrade's name
        matched = _TOKEN_RP.match(header_value, index)
        if matched is None:
            raise ValueError(f'Expected upgrade type since index {index}.')
        name = matched.group(0)
        index = matched.end()

        #nice
        result.append(name)
        
        #are we at the end?
        if index == limit:
            return result

        #lets parse till next character
        matched = _SPACE_RP.match(header_value, index)
        index = matched.end()

        #are we at the end?
        if index == limit:
            return result
    
        #no sublist?
        if header_value[index] == ',':
            index += 1
            continue
        
        raise ValueError(f'Expected \',\' at index {index}.')

def build_subprotocols(subprotocols):
    """
    Builds websocket subprotocol headers from the given subprotocol values.
    
    Parameters
    ----------
    subprotocols : `list` of `str`
        A list of supported subprotocols.
    
    Returns
    -------
    header_value : `str`
    """
    return ', '.join(subprotocols)

parse_subprotocols = parse_connections # yes, these are the same

def parse_upgrades(header_value):
    """
    Parses upgrade headers.
    
    Parameters
    ----------
    header_value : `str`
        Received upgrade header.
    
    Returns
    -------
    result : `list` of `str`
        The parsed upgrade headers.
    
    Raises
    ------
    ValueError
        Upgrade header value is incorrect.
    """
    result = []
    limit = len(header_value)
    index = 0
    
    while True:
        # parse till 1st element
        matched = _LIST_START_RP.match(header_value, index)
        index = matched.end()
        
        # are we at the end?
        if index == limit:
            return result
        
        # now lets parse the upgrade's name
        matched = _PROTOCOL_RP.match(header_value, index)
        if matched is None:
            raise ValueError(f'Expected upgrade type since index {index}.')
        name = matched.group(0)
        index = matched.end()
        
        # nice
        result.append(name)
        
        # are we at the end?
        if index == limit:
            return result
        
        # lets parse till next character
        matched = _SPACE_RP.match(header_value, index)
        index = matched.end()
        
        # are we at the end?
        if index == limit:
            return result
        
        # no sublist?
        if header_value[index] == ',':
            index += 1
            continue
        
        raise ValueError(f'Expected \',\' at index {index}.')



__all__ = ('ChannelGuildBase', 'ChannelGuildMainBase')

import re, warnings

from ...backend.utils import copy_docs
from ...backend.export import export, include

from ..permission import Permission, PermissionOverwriteTargetType
from ..permission.permission import PERMISSION_NONE, PERMISSION_ALL, PERMISSION_MASK_ADMINISTRATOR, \
    PERMISSION_MASK_VIEW_CHANNEL
from ..core import GUILDS, CHANNELS
from ..user import ClientUserBase

PERMISSION_OVERWRITE_TYPE_ROLE = PermissionOverwriteTargetType.role
PERMISSION_OVERWRITE_TYPE_USER = PermissionOverwriteTargetType.user

from .channel_base import ChannelBase

Client = include('Client')
create_partial_guild_from_id = include('create_partial_guild_from_id')

@export
class ChannelGuildBase(ChannelBase):
    """
    Base class for guild channels.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _permission_cache : `None` or `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    guild_id : `int`
        The channel's guild's identifier.
    name : `str`
        The channel's name.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `0`
        The default type, what the channel represents.
    INTERCHANGE : `tuple` of `int` = `()`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    ORDER_GROUP : `int` = `0`
        An order group what defined which guild channel type comes after the other one.
    """
    __slots__ = ('_permission_cache', 'parent_id', 'guild_id', 'name', )
    
    ORDER_GROUP = 0
    
    @copy_docs(ChannelBase._get_processed_name)
    def _get_processed_name(self):
        return self.name
    
    
    @property
    @copy_docs(ChannelBase.users)
    def users(self):
        return list(self.iter_users())
    
    
    @copy_docs(ChannelBase.iter_users)
    def iter_users(self):
        guild = self.guild
        if (guild is not None):
            for user in guild.users.values():
                if self.permissions_for(user)&PERMISSION_MASK_VIEW_CHANNEL:
                    yield user
    
    
    @property
    @copy_docs(ChannelBase.clients)
    def clients(self):
        guild = self.guild
        if guild is None:
            return []
        
        return guild.clients
    
    
    @copy_docs(ChannelBase.get_user)
    def get_user(self, name, default=None):
        name_length = len(name)
        if (name_length < 1) or (name_length > 32):
            return
        
        if self.guild is None:
            return default
        
        users = self.users
        
        if (name_length > 6) and (name[-5] == '#'):
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if (user.discriminator == discriminator) and (user.name == name_):
                        return user
        
        if name_length > 32:
            return default

        for user in users:
            if user.name == name:
                return user
        
        guild_id = self.guild_id
        
        for user in users:
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            nick = guild_profile.nick
            if nick is None:
                continue
            
            if nick == name:
                return user
        
        return default
    
    
    @copy_docs(ChannelBase.get_user_like)
    def get_user_like(self, name, default=None):
        name_length = len(name)
        if (name_length < 1) or (name_length > 37):
            return
        
        guild = self.guild
        if guild is None:
            return default
        
        if (name_length > 6) and (name[-5] == '#'):
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in guild.users.values():
                    if not self.permissions_for(user)&PERMISSION_MASK_VIEW_CHANNEL:
                        continue
                    
                    if (user.discriminator == discriminator) and (user.name == name_):
                        return user
        
        if (name_length > 32):
            return default
        
        pattern = re.compile(re.escape(name), re.I)
        
        guild_id = guild.id
        
        for user in guild.users.values():
            if not self.permissions_for(user)&PERMISSION_MASK_VIEW_CHANNEL:
                continue
            
            if pattern.match(user.name) is not None:
                return user
            
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            nick = guild_profile.nick
            if nick is None:
                continue
            
            if pattern.match(nick) is None:
                continue
            
            return user
        
        return default
    
    
    @copy_docs(ChannelBase.get_users_like)
    def get_users_like(self, name):
        result = []
        
        name_length = len(name)
        if (name_length < 1) or (name_length > 37):
            return result
        
        guild = self.guild
        if guild is None:
            return result
        
        users = self.users
        
        if (name_length > 6) and (name[-5] == '#'):
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if not self.permissions_for(user)&PERMISSION_MASK_VIEW_CHANNEL:
                        continue
                    
                    if (user.discriminator == discriminator) and (user.name == name_):
                        result.append(user)
                        break
        
        if name_length > 32:
            return result
        
        pattern = re.compile(re.escape(name), re.I)
        
        guild_id = guild.id
        for user in guild.users.values():
            if not self.permissions_for(user)&PERMISSION_MASK_VIEW_CHANNEL:
                continue
            
            if pattern.match(user.name) is not None:
                result.append(user)
                continue
            
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            nick = guild_profile.nick
            if nick is None:
                continue
            
            if pattern.match(nick) is None:
                continue
            
            result.append(user)
        
        return result
    
    
    @classmethod
    @copy_docs(ChannelBase._from_partial_data)
    def _from_partial_data(cls, data, channel_id, guild_id):
        self = super(ChannelGuildBase, cls)._from_partial_data(data, channel_id, guild_id)
        
        try:
            name = data['name']
        except KeyError:
            pass
        else:
            self.name = name
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelGuildBase, cls)._create_empty(channel_id, channel_type, guild_id)
        self._permission_cache = None
        self.parent_id = 0
        self.guild_id = guild_id
        self.name = ''
        return self
    
    
    @property
    def guild(self):
        """
        Returns the channel's guild.
        
        Returns
        -------
        guild : ``Guild`` or `None`
        """
        guild_id = self.guild_id
        if guild_id:
            return create_partial_guild_from_id(guild_id)
    
    
    @property
    def parent(self):
        """
        Returns the channel's parent channel.
        
        Returns
        -------
        parent : `None` or ``ChannelGuildBase``
        """
        parent_id = self.parent_id
        if parent_id:
            return CHANNELS.get(parent_id, None)
    
    
    @copy_docs(ChannelBase._delete)
    def _delete(self):
        self.permission_overwrites.clear()
        self._permission_cache = None


    @copy_docs(ChannelBase.cached_permissions_for)
    def cached_permissions_for(self, user):
        if not isinstance(user, Client):
            return self.permissions_for(user)
        
        permission_cache = self._permission_cache
        if permission_cache is None:
            self._permission_cache = permission_cache = {}
        else:
            try:
                return permission_cache[user.id]
            except KeyError:
                pass
        
        permissions = self.permissions_for(user)
        permission_cache[user.id] = permissions
        return permissions


@export
class ChannelGuildMainBase(ChannelGuildBase):
    """
    Base class for main guild channels not including thread channels.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    parent_id : `int`
        The channel's parent's identifier.
    guild_id : `int`
        The channel's guild's identifier.
    name : `str`
        The channel's name.
    _permission_cache : `None` or `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `0`
        The default type, what the channel represents.
    INTERCHANGE : `tuple` of `int` = `()`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    ORDER_GROUP : `int` = `0`
        An order group what defined which guild channel type comes after the other one.
    """
    __slots__ = ('permission_overwrites', 'position', )
    
    def __gt__(self, other):
        """
        Whether this channel has higher (visible) position than an other one in a guild.
        If the other channel is not guild channel, then just compares their id.
        """
        if isinstance(other, ChannelGuildBase):
            if self.ORDER_GROUP > other.ORDER_GROUP:
                return True
            
            if self.ORDER_GROUP == other.ORDER_GROUP:
                if self.position > other.position:
                    return True
                
                if self.position == other.position:
                    if self.id > other.id:
                        return True
            
            return False
        
        if isinstance(other, ChannelBase):
            return self.id > other.id
        
        return NotImplemented
    
    def __ge__(self, other):
        """
        Whether this channel is same as the other one, or has higher (visible) position than an other one in a guild.
        If the other channel is not guild channel, then just compares their id.
        """
        if isinstance(other, ChannelGuildBase):
            if self.ORDER_GROUP > other.ORDER_GROUP:
                return True
            
            if self.ORDER_GROUP == other.ORDER_GROUP:
                if self.position > other.position:
                    return True
                
                if self.position == other.position:
                    if self.id >= other.id:
                        return True
            
            return False
        
        if isinstance(other, ChannelBase):
            return self.id > other.id
        
        return NotImplemented
    
    def __le__(self, other):
        """
        Whether this channel is same as the other one, or has lower (visible) position than an other one in a guild.
        If the other channel is not guild channel, then just compares their id.
        """
        if isinstance(other, ChannelGuildBase):
            if self.ORDER_GROUP < other.ORDER_GROUP:
                return True
            
            if self.ORDER_GROUP == other.ORDER_GROUP:
                if self.position < other.position:
                    return True
                
                if self.position == other.position:
                    if self.id <= other.id:
                        return True
            
            return False
        
        if isinstance(other, ChannelBase):
            return self.id < other.id
        
        return NotImplemented
    
    def __lt__(self, other):
        """
        Whether this channel has lower (visible) position than an other one in a guild.
        If the other channel is not guild channel, then just compares their id.
        """
        if isinstance(other, ChannelGuildBase):
            if self.ORDER_GROUP < other.ORDER_GROUP:
                return True
            
            if self.ORDER_GROUP == other.ORDER_GROUP:
                if self.position < other.position:
                    return True
                
                if self.position == other.position:
                    if self.id < other.id:
                        return True
            
            return False
        
        if isinstance(other, ChannelBase):
            return self.id < other.id
        
        return NotImplemented
    
    
    def _init_parent_and_position(self, data, guild_id):
        """
        Initializes the `.parent` and the `.position` of the channel. If a channel is under the ``Guild``,
        and not in a parent (parent channels are all like these), then their `.parent` is the ``Guild`` itself.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        guild : ``Guild``
            The guild of the channel.
        """
        self.guild_id = guild_id
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            guild.channels[self.id] = self
        
        parent_id = data.get('parent_id', None)
        if (parent_id is None):
            parent_id = 0
        else:
            parent_id = int(parent_id)
        
        self.parent_id = parent_id
        
        self.position = data.get('position', 0)
    
    
    def _set_parent_and_position(self, data):
        """
        Similar to the ``._init_parent_and_position`` method, but this method applies the changes too, so moves the channel
        between categories and moves the channel inside of the parent too, to keep the order.
        
        Called from `._update_no_return` when updating a guild channel.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord
        """
        new_parent_id = data.get('parent_id', None)
        if new_parent_id is None:
            new_parent_id = 0
        else:
            new_parent_id = int(new_parent_id)
        
        position = data.get('position', 0)
        
        parent_id = self.parent_id
        if parent_id is new_parent_id:
            if self.position != position:
                self.position = position
        
        else:
            self.position = position
            self.parent_id = new_parent_id
    
    
    def _update_parent_and_position(self, data, old_attributes):
        """
        Acts same as ``._set_parent_and_position``, but it sets the modified attributes' previous value to
        `old_attributes`.
        
        Called from `._update` when updating a guild channel.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord
        old_attributes : `dict` of (`str`, `Any`) items
            `attribute-name` - `old-value` relations containing the changes of caused by the update.
        
            The method may add the following items to the dictionary:
            +---------------+-----------------------------------+
            | Keys          | Values                            |
            +===============+===================================+
            | parent_id     | `int`                             |
            +---------------+-----------------------------------+
            | position      | `int`                             |
            +---------------+-----------------------------------+
        """
        new_parent_id = data.get('parent_id', None)
        if new_parent_id is None:
            new_parent_id = 0
        else:
            new_parent_id = int(new_parent_id)
        
        position = data.get('position', 0)
        
        parent_id = self.parent_id
        if parent_id is new_parent_id:
            if self.position != position:
                old_attributes['position'] = self.position
                self.position = position
        else:
            old_attributes['parent_id'] = parent_id
            old_attributes['position'] = self.position
            
            self.position = position
            self.parent_id = parent_id
    
    
    def _permissions_for(self, user):
        """
        Base permission calculator method. Subclasses call this first, then apply their channel type related changes.
        
        Parameters
        ----------
        user : ``UserBase`` instance
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        guild_id = self.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        if user.id == guild.owner_id:
            return PERMISSION_ALL
        
        if not isinstance(user, ClientUserBase):
            if  user.guild_id in guild.channels:
                role_everyone = guild.roles.get(guild_id, None)
                if role_everyone is None:
                    permissions = PERMISSION_NONE
                else:
                    permissions = role_everyone.permissions
                
                # Apply everyone's if applicable
                permission_overwrites = self.permission_overwrites
                try:
                    permission_overwrite_everyone = permission_overwrites[guild_id]
                except KeyError:
                    pass
                else:
                    permissions &= ~permission_overwrite_everyone.deny
                    permissions |= permission_overwrite_everyone.allow
            
                return permissions
            else:
                return PERMISSION_NONE
        
        # Are we in the guild?
        try:
            guild_profile = user.guild_profiles[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        # Apply everyone's
        role_everyone = guild.roles.get(guild_id, None)
        if role_everyone is None:
            permissions = 0
        else:
            permissions = role_everyone.permissions
        
        roles = guild_profile.roles
        
        # Apply permissions of each role of the user.
        if (roles is not None):
            for role in roles:
                permissions |= role.permissions
        
        # Apply everyone's if applicable
        permission_overwrites = self.permission_overwrites
        try:
            permission_overwrite_everyone = permission_overwrites[guild_id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_everyone.deny
            permissions |= permission_overwrite_everyone.allow
        
        # Apply role overwrite
        allow = 0
        deny = 0
        
        if (roles is not None):
            for role in roles:
                try:
                    permission_overwrite_role = permission_overwrites[role.id]
                except KeyError:
                    pass
                else:
                    allow |= permission_overwrite_role.allow
                    deny |= permission_overwrite_role.deny
        
        permissions &= ~deny
        permissions |= allow
        
        # Apply user specific
        try:
            permission_overwrite_user = permission_overwrites[user.id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_user.deny
            permissions |= permission_overwrite_user.allow
        
        # Are we admin?
        if permissions&PERMISSION_MASK_ADMINISTRATOR:
            return PERMISSION_ALL
        
        return Permission(permissions)
    
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self, user):
        result = self._permissions_for(user)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            result = PERMISSION_NONE
        
        return result
    
    
    def _permissions_for_roles(self, roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles. This method is called
        first by subclasses to apply their own related permissions on it.
        
        Parameters
        ----------
        roles : `tuple` of ``Role``
            The roles to calculate final permissions from.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        
        Notes
        -----
        Partial roles and roles from other guilds as well are ignored.
        """
        guild = self.guild
        if guild is None:
            return PERMISSION_NONE
        
        guild_id = guild.id
        role_everyone = guild.roles.get(guild_id, None)
        if role_everyone is None:
            permissions = 0
        else:
            permissions = role_everyone.permissions
        
        roles = sorted(roles)
        for role in roles:
            if role.guild_id == guild_id:
                permissions |= role.permissions
        
        if permissions&PERMISSION_MASK_ADMINISTRATOR:
            return PERMISSION_ALL
        
        permission_overwrites = self.permission_overwrites
        try:
            permission_overwrite_everyone = permission_overwrites[guild_id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_everyone.deny
            permissions |= permission_overwrite_everyone.allow
        
        allow = 0
        deny = 0
        
        for role in roles:
            try:
                permission_overwrite_role = permission_overwrites[role.id]
            except KeyError:
                pass
            else:
                allow |= permission_overwrite_role.allow
                deny |= permission_overwrite_role.deny
        
        permissions &= ~deny
        permissions |= allow
        
        return Permission(permissions)
    
    
    @copy_docs(ChannelBase.permissions_for_roles)
    def permissions_for_roles(self, *roles):
        permissions = self._permissions_for_roles(roles)
        if not permissions&PERMISSION_MASK_VIEW_CHANNEL:
            permissions = PERMISSION_NONE
        
        return permissions
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelGuildMainBase, cls)._create_empty(channel_id, channel_type, guild_id)
        self.permission_overwrites = {}
        self.position = 0
        return self
    
    
    @copy_docs(ChannelBase._delete)
    def _delete(self):
        self.permission_overwrites.clear()
        self._permission_cache = None
        
        try:
            guild = GUILDS[self.guild_id]
        except KeyError:
            pass
        else:
            try:
                del guild.channels[self.id]
            except KeyError:
                pass



__all__ = ('ChannelBase',)

import re

from ...backend.export import export, include

from ..bases import DiscordEntity
from ..permission import Permission
from ..permission.permission import PERMISSION_NONE
from ..user import User
from ..utils import DATETIME_FORMAT_CODE

Client = include('Client')

@export
class ChannelBase(DiscordEntity, immortal=True):
    """
    Base class for Discord channels.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `0`
        The default type, what the channel represents.
    INTERCHANGE : `tuple` of `int` = `()`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    
    Notes
    -----
    Channels support weakreferencing.
    """
    DEFAULT_TYPE = 0
    INTERCHANGE = ()
    
    def __new__(cls, data, client, guild_id):
        """
        Creates a new channel from the channel data received from Discord. If the channel already exists and if it
        is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        client : `None` or ``Client``
            The client, who received the channel's data, if any.
        guild_id : `int`
            The guild's identifier of the channel.
        
        Raises
        -------
        RuntimeError
            The respective channel type cannot be instanced.
        """
        raise RuntimeError(f'`{cls.__name__}` cannot be instanced.')
    
    
    def __repr__(self):
        """Returns the representation of the channel."""
        return f'<{self.__class__.__name__} id={self.id}, name={self._get_processed_name()!r}>'
    
    
    def _get_processed_name(self):
        """Returns the channel's name."""
        return ''
    
    
    def __format__(self, code):
        """
        Formats the channel in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        channel : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import ChannelText, now_as_id
        >>> channel = ChannelText.precreate(now_as_id(), name='GENERAL')
        >>> channel
        <ChannelText id=710506058560307200, name='GENERAL'>
        >>> # no code stands for `channel.name`.
        >>> f'{channel}'
        'GENERAL'
        >>> # 'd' stands for display name.
        >>> f'{channel:d}'
        'general'
        >>> # 'm' stands for mention.
        >>> f'{channel:m}'
        '<#710506058560307200>'
        >>> # 'c' stands for created at.
        >>> f'{channel:c}'
        '2020.05.14-14:57:24'
        ```
        """
        if not code:
            return self._get_processed_name()
        
        if code == 'm':
            return self.mention
        
        if code == 'd':
            return self.display_name
        
        if code == 'c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    @property
    def display_name(self):
        """
        The channel's display name.
        
        Returns
        -------
        display_name : `str`
        """
        return ''
    
    @property
    def mention(self):
        """
        The channel's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'<#{self.id}>'
    
    @property
    def partial(self):
        """
        Whether this channel is partial.
        
        A channel is partial if non of the running clients can see it.
        
        Returns
        -------
        is_partial : `bool`
        """
        return (not self.clients)
    
    def get_user(self, name, default=None):
        """
        Tries to find the a user with the given name at the channel. Returns the first matched one.
        
        Parameters
        ----------
        name : `str`
            The name to search for.
        default : `Any`, Optional
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase`` or `None`
        """
        if (not 1 < len(name) < 38):
            return default
        
        users = self.users
        
        if len(name) > 6 and name[-5] == '#':
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if user.discriminator == discriminator and user.name == name_:
                        return user
        
        if len(name) > 32:
            return default
        
        for user in users:
            if user.name == name:
                return user
        
        return default
    
    def get_user_like(self, name, default=None):
        """
        Searches a user, who's name or nick starts with the given string and returns the first find.
        
        Parameters
        ----------
        name : `str`
            The name to search for.
        default : `Any`, Optional
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase`` or `default`
        """
        if (not 1 < len(name) < 38):
            return default
        
        users = self.users
        
        if len(name) > 6 and name[-5] == '#':
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if user.discriminator == discriminator and user.name == name_:
                        return user
        
        if len(name) > 32:
            return default
        
        pattern = re.compile(re.escape(name), re.I)
        for user in users:
            if pattern.match(user.name) is None:
                continue
            
            return user
        
        return default
    
    def get_users_like(self, name):
        """
        Searches the users, who's name or nick starts with the given string.
        
        Parameters
        ----------
        name : `str`
            The name to search for.
        
        Returns
        -------
        users : `list` of ``ClientUserBase`` objects
        """
        result = []
        if (not 1 < len(name) < 38):
            return result
        
        users = self.users
        
        if len(name) > 6 and name[-5] == '#':
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if user.discriminator == discriminator and user.name == name_:
                        result.append(user)
                        break
        
        if len(name) > 32:
            return result
        
        pattern = re.compile(re.escape(name), re.I)
        for user in users:
            if pattern.match(user.name) is None:
                continue
            
            result.append(user)
        
        return result
    
    @property
    def users(self):
        """
        The users who can see this channel.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return []
    
    def iter_users(self):
        """
        Iterates over the users who can see the channel.
        
        This method is a generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        yield from self.users
    
    @property
    def clients(self):
        """
        The clients, who can access this channel.
        
        Returns
        -------
        clients : `list` of ``Client``
        """
        clients = []
        for user in self.users:
            if isinstance(user, Client):
                clients.append(user)
        
        return clients
    
    # for sorting channels
    def __gt__(self, other):
        """Returns whether this channel's id is greater than the other's."""
        if not isinstance(other, ChannelBase):
            return NotImplemented
        
        return self.id > other.id
        
    
    def __ge__(self, other):
        """Returns whether this channel's id is greater or equal than the other's."""
        if not isinstance(other, ChannelBase):
            return NotImplemented
        
        return self.id >= other.id
        
    
    def __eq__(self, other):
        """Returns whether this channel's id is equal to the other's."""
        if not isinstance(other, ChannelBase):
            return NotImplemented
        
        return self.id == other.id
        
    
    def __ne__(self,other):
        """Returns whether this channel's id is not equal to the other's."""
        if not isinstance(other, ChannelBase):
            return NotImplemented
        
        return self.id != other.id
        
    
    def __le__(self, other):
        """Returns whether this channel's id is less or equal than the other's."""
        if not isinstance(other, ChannelBase):
            return NotImplemented
        
        return self.id <= other.id
        
    
    def __lt__(self, other):
        """Returns whether this channel's id is less than the other's."""
        if not isinstance(other, ChannelBase):
            return NotImplemented
        
        return self.id < other.id
        
    
    @property
    def name(self):
        """
        Returns the channel's name.
        
        Subclasses should overwrite it.
        
        Returns
        -------
        name : `str`
        """
        return self.__class__.__name__
    
    
    def has_name_like(self, name):
        """
        Returns whether the channel's name is like the given string.
        
        Parameters
        ----------
        name : `str`
            The name of the channel
        
        Returns
        -------
        channel : ``ChannelBase`` instance
        """
        if name.startswith('#'):
            name = name[1:]
        
        target_name_length = len(name)
        if (target_name_length < 2) or (target_name_length > 100):
            return False
        
        if re.match(re.escape(name), self.name, re.I) is None:
            return False
        
        return True
    
    
    def permissions_for(self, user):
        """
        Returns the permissions for the given user at the channel.
        
        Parameters
        ----------
        user : ``UserBase`` instance
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
            
        
        See Also
        --------
        ``.cached_permissions_for`` : Cached permission calculator.
        
        Notes
        -----
        Always return empty permissions. Subclasses should implement this method.
        """
        return PERMISSION_NONE
    
    
    def cached_permissions_for(self, user):
        """
        Returns the permissions for the given user at the channel. If the user's permissions are not cached, calculates
        and stores them first.
        
        Parameters
        ----------
        user : ``UserBase`` instance
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        
        Notes
        -----
        Mainly designed for getting clients' permissions and stores only their as well. Do not caches other user's
        permissions.
        
        Always return empty permissions. Subclasses should implement this method.
        """
        return PERMISSION_NONE
    
    
    def permissions_for_roles(self, *roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles.
        
        Parameters
        ----------
        *roles : ``Role``
            The roles to calculate final permissions from.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        
        Notes
        -----
        Partial roles and roles from other guilds as well are ignored.
        
        Always return empty permissions. Subclasses should implement this method.
        """
        return PERMISSION_NONE
    
    
    @property
    def guild_id(self):
        """
        Returns the channel's guild's identifier. At the case of private channels, this always returns `0`.
        
        Returns
        -------
        channel_id : `int`
        """
        return 0
    
    @property
    def guild(self):
        """
        Returns the channel's guild. At the case of private channels this is always `None`.
        
        Returns
        -------
        guild : `None`
        """
        return None
    
    
    def _update_attributes(self, data):
        """
        Updates the channel with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        """
        pass
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        """
        return {}
    
    
    def _delete(self):
        """
        Removes the channel's references.
        
        Used when the channel is deleted.
        """
        pass

    
    @classmethod
    def _from_partial_data(cls, data, channel_id, guild_id):
        """
        Creates a channel from partial data. Called by ``create_partial_channel_from_data`` when a new
        partial channel is needed to be created.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Partial channel data.
        channel_id : `int`
            The channel's id.
        guild_id : `int`
            The channel's guild's identifier if applicable.
        
        Returns
        -------
        channel : ``ChannelBase``
        """
        try:
            channel_type = data['type']
        except KeyError:
            channel_type = cls.DEFAULT_TYPE
        
        self = cls._create_empty(channel_id, channel_type, guild_id)
        
        return self
    
    @classmethod
    def _create_empty(cls, channel_id, channel_type, guild_id):
        """
        Creates a partial channel from the given parameters.
        
        Parameters
        ----------
        channel_id : `int`
            The channel's identifier.
        channel_type : `int`
            The channel's type identifier.
        guild_id : `int`
            A partial guild's identifier for the created channel.
        
        Returns
        -------
        channel : ``ChannelBase`` instance
            The created partial channel.
        """
        self = object.__new__(cls)
        self.id = channel_id
        return self




__all__ = ()

import base64, binascii, re
from collections import namedtuple
from urllib.parse import quote
import socket as module_socket

from .futures import CancelledError

CHAR = {chr(i) for i in range(0, 128)}
CTL = {*(chr(i) for i in range(0, 32)), chr(127)}
SEPARATORS = {'(', ')', '<', '>', '@', ', ', ';', ':', '\\', '"', '/', '[', ']', '?', '=', '{', '}', ' ', chr(9)}
TOKEN = CHAR ^ CTL ^ SEPARATORS

sentinel = object()

HttpVersion = namedtuple('HttpVersion', ('major', 'minor'))
HttpVersion10 = HttpVersion(1,0)
HttpVersion11 = HttpVersion(1,1)

del namedtuple

class BasicAuth:
    """
    Http basic authorization implementation.
    
    Attributes
    ----------
    username : `str`
        Authorization login username.
    password : `str`
        Authorization password. Can be empty string.
    encoding : `str`
        Encoding used to encode the authorization headers. Defaults to ``.DEFAULT_ENCODING``.
    
    Class Attributes
    ----------------
    DEFAULT_ENCODING : `str` = `'latin1'`
        Default encoding used to encode and decode the authorization headers.
    """
    DEFAULT_ENCODING = 'latin1'
    __slots__ = ('username', 'password', 'encoding',)
    
    def __new__(cls, username, password='', encoding=DEFAULT_ENCODING):
        """
        Creates a new ``BasicAuth`` instance with the given parameters.
        
        Attributes
        ----------
        username : `str`
            Authorization login name.
        password : `str`, Optional
            Authorization password. Can be empty string.
        encoding : `str`, Optional
            Encoding used to encode the authorization headers. Defaults to `'latin1'`.
        
        Raises
        ------
        ValueError
            If `username` is given as `None`.
            If `username` contains `':'` character.
            If password is given as `None`.
        """
        if username is None:
            raise ValueError('`None` is not allowed as `login` value.')
        
        if ':' in username:
            raise ValueError(f'`\':\'` is not allowed in `login` (RFC 1945#section-11.1), got {username!r}.')
        
        if password is None:
            raise ValueError('`None` is not allowed as `password` value.')
        
        self = object.__new__(cls)
        
        self.username = username
        self.password = password
        self.encoding = encoding
        
        return self
    
    @classmethod
    def decode(cls, auth_header, encoding=DEFAULT_ENCODING):
        """
        Creates a new ``BasicAuth`` instance from the given HTTP header value and.
        
        Parameters
        ----------
        auth_header : `str`
            Authorization header value.
        encoding : `str`, Optional
            Encoding used to encode the authorization headers. Defaults to `'latin1'`.
        
        Returns
        -------
        self : ``BasicAuth``.
        
        Raises
        ------
        ValueError
            If the authorization method is not `'basic'`.
            If cannot parse the authorization headers.
            Cannot decode authorization header.
        """
        split = auth_header.strip().split(' ')
        if len(split) == 2:
            if split[0].strip().lower() != 'basic':
                raise ValueError(f'Unknown authorization method: {split[0]!r}.')
            to_decode = split[1]
        else:
            raise ValueError(f'Could not parse authorization header from: {auth_header!r}.')
        
        try:
            username, _, password = base64.b64decode(to_decode.encode('ascii')).decode(encoding).partition(':')
        except binascii.Error as err:
            raise ValueError('Invalid base64 encoding.') from err
        
        self = object.__new__(cls)
        self.username = username
        self.password = password
        self.encoding = encoding
        return self
    
    def encode(self):
        """
        Encodes the authorization to it's header value.
        
        Returns
        -------
        auth_header : `str`
        """
        credits_ = (f'{self.username}:{self.password}').encode(self.encoding)
        sub_value = base64.b64encode(credits_).decode(self.encoding)
        return f'Basic {sub_value}'
    
    def __repr__(self):
        """Returns the basic auth's representation."""
        result = [
            self.__class__.__name__,
            '(username=',
            repr(self.username),
                ]
        
        password = self.password
        if password:
            result.append(', password=')
            result.append(repr(password))
        
        encoding = self.encoding
        if encoding != self.DEFAULT_ENCODING:
            result.append(' encoding=')
            result.append(repr(encoding))
        
        result.append(')')
        
        return ''.join(result)

_ipv4_pattern = '^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'
_ipv6_pattern = (
    '^(?:(?:(?:[A-F0-9]{1,4}:){6}|(?=(?:[A-F0-9]{0,4}:){0,6}'
    '(?:[0-9]{1,3}\.){3}[0-9]{1,3}$)(([0-9A-F]{1,4}:){0,5}|:)'
    '((:[0-9A-F]{1,4}){1,5}:|:)|::(?:[A-F0-9]{1,4}:){5})'
    '(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}'
    '(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])|(?:[A-F0-9]{1,4}:){7}'
    '[A-F0-9]{1,4}|(?=(?:[A-F0-9]{0,4}:){0,7}[A-F0-9]{0,4}$)'
    '(([0-9A-F]{1,4}:){1,7}|:)((:[0-9A-F]{1,4}){1,7}|:)|(?:[A-F0-9]{1,4}:){7}'
    ':|:(:[A-F0-9]{1,4}){7})$'
        )

_ipv4_regex = re.compile(_ipv4_pattern)
_ipv6_regex = re.compile(_ipv6_pattern, flags=re.I)
_ipv4_regex_b = re.compile(_ipv4_pattern.encode('ascii'))
_ipv6_regex_b = re.compile(_ipv6_pattern.encode('ascii'), flags=re.I)

del _ipv4_pattern, _ipv6_pattern, re

def is_ip_address(host):
    """
    Returns whether the given `host` is an ip address.
    
    Parameters
    ----------
    host : `str` or `bytes-like`
        Host value.
    
    Returns
    -------
    is_ip_address : `bool`
    
    Raises
    ------
    TypeError
        If `host` was not given neither as `str`, or as `bytes-like`.
    """
    if host is None:
        return False
    
    if isinstance(host, str):
        if _ipv4_regex.match(host) is not None:
            return True
        
        if _ipv6_regex.match(host) is not None:
            return True
        
        return False
        
    if isinstance(host, (bytes, bytearray, memoryview)):
        if _ipv4_regex_b.match(host) is not None:
            return True
        
        if _ipv6_regex_b.match(host) is not None:
            return True
        
        return False
    
    raise TypeError(f'`host` should be given as `str` or as `bytes-like`, got {host.__class__.__name__}.')


TIMEOUT_STATE_NONE = 0
TIMEOUT_STATE_TIMEOUTED = 1
TIMEOUT_STATE_CANCELLED = 2
TIMEOUT_STATE_EXITED = 3

class Timeout:
    """
    Implements async timeout feature as a context manager.
    
    Attributes
    ----------
    _handle : `None` or ``TimerHandle``
        Timer handle to cancel the respective task when it occurs. After timeout occurs, or the timeouter is
        cancelled, set as `None`.
    _loop : ``EventThread``
        The event loop to what the timeouter is bound to.
    _task : `None` or ``Task``
        The respective task what will be cancelled. Set as `None` at creation and when cancelled or exited.
    _state : `int`
        The timeouter's state.
        
        Can be one of the following:
        +---------------------------+-------+
        | Respective name           | Value |
        +===========================+=======+
        | TIMEOUT_STATE_NONE        | 0     |
        +---------------------------+-------+
        | TIMEOUT_STATE_TIMEOUTED   | 1     |
        +---------------------------+-------+
        | TIMEOUT_STATE_CANCELLED   | 2     |
        +---------------------------+-------+
        | TIMEOUT_STATE_EXITED      | 3     |
        +---------------------------+-------+
    """
    __slots__ = ('_handle', '_loop', '_task', '_state')
    def __new__(cls, loop, timeout):
        """
        Creates a new timeouter instance bound to the given loop.
        
        The timeout starts when the timeouter is created.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the timeouter will be bound to.
        timeout : `float`
            Time in seconds after the task is cancelled. When the cancellation reaches the context manager, raises
            `TimeoutError` instead.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._handle = loop.call_later(timeout, cls._timeout, self)
        self._task = None
        self._state = TIMEOUT_STATE_NONE
        return self
    
    def cancel(self):
        """
        Cancels the timeouter.
        
        If already cancelled, does nothing.
        """
        if self._state == TIMEOUT_STATE_CANCELLED:
            return
        
        self._state = TIMEOUT_STATE_CANCELLED
        
        handle = self._handle
        if (handle is not None):
            self._handle = None
            handle.cancel()
        
        self._task = None
    
    def __enter__(self):
        """
        Enters the timeouter as a context manager.
        
        Returns
        -------
        self : ``Timeout``
        
        Raises
        ------
        RuntimeError
            ``Timeout`` entered outside of a ``Task``.
        TimeoutError
            Timeout already occurred.
        """
        task = self._loop.current_task
        if (task is None):
            raise RuntimeError('`Timeout` entered outside of a `Task`!')
        
        state = self._state
        if state == TIMEOUT_STATE_NONE:
            self._task = task
        elif state == TIMEOUT_STATE_TIMEOUTED:
            raise TimeoutError
        elif state == TIMEOUT_STATE_CANCELLED:
            pass
        else:
            raise RuntimeError('`Timeout` already used up.')
        
        return self
    
    def _timeout(self):
        """
        Timeouts the timeouter if not yet timeouted nor cancelled.
        """
        if self._state != TIMEOUT_STATE_NONE:
            return
        
        self._state = TIMEOUT_STATE_TIMEOUTED
        
        handle = self._handle
        if (handle is not None):
            self._handle = None
        
        task = self._task
        if (task is not None):
            task.cancel()
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exits from the timeouter. If the timeout occurs, then raises ``TimeoutError`` from the received cancellation.
        """
        handle = self._handle
        if (handle is not None):
            self._handle = None
            handle.cancel()
        
        self._task = None
        
        state = self._state
        self._state = TIMEOUT_STATE_EXITED
        if (state == TIMEOUT_STATE_TIMEOUTED) and (exc_type is CancelledError):
            raise TimeoutError from None
        
        return False
    
    def __repr__(self):
        return f'<{self.__class__.__name__}>'


def content_disposition_header(disposition_type, parameters, quote_fields=True):
    """
    Creates Content-Disposition header value.
    
    Parameters
    ----------
    disposition_type : `str`
        Disposition type. Can be one of following: `'inline'`, `'attachment'`, '`form-data`'.
    parameters : `dict` of (`str`, `str`) items
        Disposition parameters.
    quote_fields : `bool`
        Whether field values should be quoted.
    
    Returns
    -------
    value : `str`
    """
    if not disposition_type or not (TOKEN > set(disposition_type)):
        raise ValueError(f'Bad content disposition type {disposition_type!r}.')
    
    if parameters:
        parameter_parts = [disposition_type]
        for key, value in parameters.items():
            if (not key) or (not (TOKEN > set(key))):
                raise ValueError(f'Bad content disposition parameter {key!r}={value!r}.')
            
            if quote_fields:
                value = quote(value, '')
            
            parameter_parts.append(f'{key}="{value}"')
            
            if key == 'filename':
                parameter_parts.append(f'filename*=utf-8\'\'{value}')
        
        value = '; '.join(parameter_parts)
    else:
        value = disposition_type
    
    return value

def tcp_nodelay(transport, value):
    """
    Sets or removes tcp nodelay socket option to the given transport's socket if applicable.
    
    Parameters
    ----------
    transport : `Any`
        Asynchronous transport implementation.
    value : `bool`
        Value to set tcp nodelay to.
    """
    socket = transport.get_extra_info('socket')
    
    if socket is None:
        return
    
    if socket.family not in (module_socket.AF_INET, module_socket.AF_INET6):
        return
    
    # socket may be closed already, on windows OSError get raised
    try:
        socket.setsockopt(module_socket.IPPROTO_TCP, module_socket.TCP_NODELAY, value)
    except OSError:
        pass



__all__ = ()

from math import inf
from datetime import datetime
from base64 import b64decode
from binascii import Error as Base64DecodeError

from ...backend.utils import basemethod
from ...backend.event_loop import LOOP_TIME
from ...backend.futures import Future, Task, WaitTillFirst

from ..core import KOKORO, CLIENTS, CHANNELS
from ..http import RateLimitProxy
from ..utils import time_now, DISCORD_EPOCH
from ..exceptions import DiscordException
from ..channel import ChannelThread
from ..user import create_partial_user_from_id, thread_user_create
from ..permission.permission import PERMISSION_MASK_VIEW_CHANNEL

USER_CHUNK_TIMEOUT = 2.5

class SingleUserChunker:
    """
    A user chunk waiter, which yields after the first received chunk. Used at ``Client.request_members``.
    
    Attributes
    ----------
    timer : `Handle` or `None`
        The time-outer of the chunker, what will cancel if the timeout occurs.
    waiter : ``Future``
        The waiter future what will yield, when we receive the response, or when the timeout occurs.
    """
    __slots__ = ('timer', 'waiter',)
    
    def __init__(self, ):
        self.waiter = Future(KOKORO)
        self.timer = KOKORO.call_at(LOOP_TIME()+USER_CHUNK_TIMEOUT, type(self)._cancel, self)
    
    def __call__(self, event):
        """
        Called when a chunk is received with it's respective nonce.
        
        Parameters
        ----------
        event : ``GuildUserChunkEvent``
            The received guild user chunk's event.
        
        Returns
        -------
        is_last : `bool`
            ``SingleUserChunker`` returns always `True`, because it waits only for one event.
        """
        self.waiter.set_result_if_pending(event.users)
        timer = self.timer
        if (timer is not None):
            self.timer = None
            timer.cancel()
        
        return True
    
    def _cancel(self):
        """
        The chunker's timer calls this method.
        
        Cancels ``.waiter`` and ``.timer``. After this method was called, the waiting coroutine will remove it's
        reference from the event handler.
        """
        self.waiter.cancel()
        
        timer = self.timer
        if (timer is not None):
            self.timer = None
            timer.cancel()
    
    def cancel(self):
        """
        Cancels the chunker.
        
        This method should be called when when the chunker is canceller from outside. Before this method is called,
        it's references should be removed as well from the event handler.
        """
        self.waiter.set_result_if_pending([])
        
        timer = self.timer
        if (timer is not None):
            self.timer = None
            timer.cancel()
    
    def __await__(self):
        """
        Awaits the chunker's waiter and returns that's result.
        
        Returns
        -------
        users : `list` of ``ClientUserBase`` objects
            The received users. Can be an empty list.
        
        Raises
        ------
        CancelledError
            If timeout occurred.
        """
        return self.waiter.__await__()


class MassUserChunker:
    """
    A user chunk waiter, which yields after the chunks of certain amount of guilds are received. Used at
    ``Client._request_members`` and at ``Client._request_members``.
    
    Attributes
    ----------
    last : `float`
        The timestamp of the last received chunk.
    timer : `Handle` or `None`
        The time-outer of the chunker, what will cancel if the timeout occurs.
    waiter : ``Future``
        The waiter future what will yield, when we receive the response, or when the timeout occurs.
    """
    __slots__ = ('last', 'timer', 'waiter',)
    
    def __init__(self, left):
        """
        Parameters
        ----------
        left : `int`
            How much guild's chunks are left to be received.
        """
        self.waiter = Future(KOKORO)
        self.last = now = LOOP_TIME()
        self.timer = KOKORO.call_at(now+USER_CHUNK_TIMEOUT, type(self)._cancel, self)
    
    def __call__(self, event):
        """
        Called when a chunk is received with it's respective nonce.
        
        Updates the chunker's last received chunk's time to push out the current timeout.
        
        Parameters
        ----------
        event : ``GuildUserChunkEvent``
            The received guild user chunk's event.
        
        Returns
        -------
        is_last : `bool`
            Whether the last chunk was received.
        """
        self.last = LOOP_TIME()
        if event.index+1 != event.count:
            return False
        
        self.waiter.set_result_if_pending(None)
        timer = self.timer
        if (timer is not None):
            self.timer = None
            timer.cancel()
        
        return True
    
    def _cancel(self):
        """
        The chunker's timer calls this method. If the chunker received any chunks since it's ``.timer`` was started,
        pushes out the timeout.
        
        Cancels ``.waiter`` and ``.timer``. After this method was called, the waiting coroutine will remove it's
        reference from the event handler.
        """
        now = LOOP_TIME()
        next_ = self.last + USER_CHUNK_TIMEOUT
        if next_ > now:
            self.timer = KOKORO.call_at(next_, type(self)._cancel, self)
        else:
            self.timer = None
            self.waiter.cancel()
    
    def cancel(self):
        """
        Cancels the chunker.
        
        This method should be called when when the chunker is canceller from outside. Before this method is called,
        it's references should be removed as well from the event handler.
        """
        self.waiter.set_result_if_pending(None)
        
        timer = self.timer
        if (timer is not None):
            self.timer = None
            timer.cancel()
    
    def __await__(self):
        """
        Awaits the chunker's waiter.
        
        Raises
        ------
        CancelledError
            If timeout occurred.
        """
        return self.waiter.__await__()


class DiscoveryCategoryRequestCacher:
    """
    Cacher for storing ``Client``'s requests.
    
    Attributes
    ----------
    _active_request : `bool`
        Whether there is an active request.
    _last_update : `float`
        The last time when the cache was updated
    _waiter : ``Future`` or `None`
        Waiter to avoid concurrent calls.
    cached : `Any`
        Last result.
    func : `callable`
        Async callable, what's yields are cached.
    timeout : `float`
        The time interval between what the requests should be done.
    """
    __slots__ = ('_active_request', '_last_update', '_waiter', 'cached', 'func', 'timeout',)
    
    def __init__(self, func, timeout, cached=...):
        """
        Creates a ``DiscoveryCategoryRequestCacher`` instance.
        
        Parameters
        ----------
        timeout : `float`
            The time after new request should be executed.
        func : `callable`
            Async callable, what's yields would be cached.
        cached : `Any`, Optional
            Whether there should be an available cache by default.
        """
        self.func = func
        self.timeout = timeout
        self.cached = cached
        self._waiter = None
        self._active_request = False
        self._last_update = -inf
    
    def __get__(self, client, type_):
        if client is None:
            return self
        
        return basemethod(self.__class__.execute, self, client)
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')
    
    async def execute(self, client):
        """
        Executes the request and returns it's result or raises.
        
        This method is a coroutine.
        
        Returns
        -------
        result : `Any`
        
        Raises
        ------
        ConnectionError
            If there is no internet connection, or there is no available cached result.
        DiscordException
            If any exception was received from the Discord API.
        """
        if (LOOP_TIME() - self.timeout) < self._last_update:
            if self._active_request:
                waiter = self._waiter
                if waiter is None:
                    waiter = self._waiter = Future(KOKORO)
                
                result = await waiter
            else:
                result = self.cached
            
            return result
        
        self._active_request = True
        try:
            result = await self.func(client)
        except ConnectionError as err:
            result = self.cached
            if (result is ...):
                waiter = self._waiter
                if (waiter is not None):
                    self._waiter = None
                    waiter.set_exception(err)
                
                raise
        
        except BaseException as err:
            waiter = self._waiter
            if (waiter is not None):
                self._waiter = None
                waiter.set_exception(err)
            
            raise
        
        else:
            self._last_update = LOOP_TIME()
        
        finally:
            self._active_request = False
        
        waiter = self._waiter
        if (waiter is not None):
            self._waiter = None
            waiter.set_result(result)
        
        return result
    
    def __repr__(self):
        """Returns the cacher's representation."""
        result = [
            self.__class__.__name__,
            '(func=',
            repr(self.func),
            ', timeout=',
            repr(self.timeout),
        ]
        
        cached = self.cached
        if (cached is not ...):
            result.append(' cached=')
            result.append(repr(cached))
        
        result.append(')')
        
        return ''.join(result)
    
    __call__ = execute


class TimedCacheUnit:
    """
    Timed cache unit used at keyed request cachers.
    
    Attributes
    ----------
    result : `str`
        The cached response object.
    creation_time : `float`
        The LOOP_TIME time when the last response was received.
    last_usage_time : `float`
        The monotonic time when this unit was last time used.
    """
    __slots__ = ('creation_time', 'last_usage_time', 'result')
    def __repr__(self):
        """Returns the timed cache unit's representation."""
        return (f'<{self.__class__.__name__} creation_time={self.creation_time!r}, last_usage_time='
                f'{self.last_usage_time!r}, result={self.result!r}>')


class DiscoveryTermRequestCacher:
    """
    Cacher for storing ``Client'' requests. Also uses other clients, if the source client's rate limits are already
    exhausted.
    
    Attributes
    ----------
    _last_cleanup : `float`
        The last time when a cleanup was done.
    _minimal_cleanup_interval : `float`
        The minimal time what needs to pass between cleanups.
    _rate_limit_proxy_args : `tuple` (``RateLimitGroup``, (``DiscordEntity`` or `None`))
        Rate limit proxy parameters used when looking up the rate limits of clients.
    _waiters : `dict` of (`str`, `
        Waiters for requests already being done.
    cached : `dict`
        Already cached responses.
    func : `callable`
        Async callable, what's yields are cached.
    timeout : `float`
        The timeout after the new request should be done instead of using the already cached response.
    """
    __slots__ =('_last_cleanup', '_minimal_cleanup_interval', '_rate_limit_proxy_args', '_waiters', 'cached', 'func',
        'timeout')
    
    def __init__(self, func, timeout, rate_limit_group, rate_limit_limiter=None,):
        """
        Creates a new ``DiscoveryTermRequestCacher`` object with the given parameters.
        
        Parameters
        ----------
        func : `callable`
            Async callable, what's yields are cached.
        timeout : `float`
            The timeout after the new request should be done instead of using the already cached response.
        rate_limit_group : ``RateLimitGroup``
            Rate limit group of the respective request.
        rate_limit_limiter : ``DiscordEntity``, Optional
            Rate limit limiter fo the respective request.
        """
        self.func = func
        self.timeout = timeout
        self.cached = {}
        self._rate_limit_proxy_args = (rate_limit_group, rate_limit_limiter)
        self._waiters = {}
        minimal_cleanup_interval = timeout / 10.0
        if minimal_cleanup_interval < 1800.0:
            minimal_cleanup_interval = 1800.0
        
        self._minimal_cleanup_interval = minimal_cleanup_interval
        self._last_cleanup = -inf
    
    def __get__(self, client, type_):
        if client is None:
            return self
        
        return basemethod(self.__class__.execute, self, client)
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')
    
    async def execute(self, client, arg):
        """
        Executes the request and returns it's result or raises.
        
        This method is a coroutine.
        
        Returns
        -------
        result : `Any`
        
        Raises
        ------
        ConnectionError
            If there is no internet connection, or there is no available cached result.
        TypeError
            The given `arg` was not passed as `str` instance.
        DiscordException
            If any exception was received from the Discord API.
        """
        # First check arg
        arg_type = arg.__class__
        if arg_type is str:
            pass
        elif issubclass(arg_type, str):
            arg = str(arg)
        else:
            raise TypeError(f'The parameter can be given as `str` instance, got {arg_type.__class__}.')
        
        # First check cache
        try:
            unit = self.cached[arg]
        except KeyError:
            unit = None
        else:
            now = LOOP_TIME()
            if self.timeout + unit.creation_time > now:
                unit.last_usage_time = now
                return unit.result
        
        # Second check actual request
        try:
            waiter = self._waiters[arg]
        except KeyError:
            pass
        else:
            if waiter is None:
                self._waiters[arg] = waiter = Future(KOKORO)
            
            return await waiter
        
        # No actual request is being done, so mark that we are doing a request.
        self._waiters[arg] = None
        
        # Search client with free rate limits.
        free_count = RateLimitProxy(client, *self._rate_limit_proxy_args).free_count
        if not free_count:
            requester = client
            for client_ in CLIENTS.values():
                if client_ is client:
                    continue
                
                free_count = RateLimitProxy(client_, *self._rate_limit_proxy_args).free_count
                if free_count:
                    requester = client_
                    break
                
                continue
            
            # If there is no client with free count do not care about the reset times, because probably only 1 client
            # forces requests anyways, so that's rate limits will reset first as well.
            client = requester
        
        # Do the request
        try:
            result = await self.func(client, arg)
        except ConnectionError as err:
            if (unit is None):
                waiter = self._waiters.pop(arg)
                if (waiter is not None):
                    waiter.set_exception(err)
                
                raise
            
            unit.last_usage_time = LOOP_TIME()
            result = unit.result
        
        except BaseException as err:
            waiter = self._waiters.pop(arg, None)
            if (waiter is not None):
                waiter.set_exception(err)
            
            raise
        
        else:
            if unit is None:
                self.cached[arg] = unit = TimedCacheUnit()
            
            now = LOOP_TIME()
            unit.last_usage_time = now
            unit.creation_time = now
            unit.result = result
        
        finally:
            # Do cleanup if needed
            now = LOOP_TIME()
            if self._last_cleanup + self._minimal_cleanup_interval < now:
                self._last_cleanup = now
                
                cleanup_till = now - self.timeout
                collected = []
                
                cached = self.cached
                for cached_arg, cached_unit in cached.items():
                    if cached_unit.last_usage_time < cleanup_till:
                        collected.append(cached_arg)
                
                for cached_arg in collected:
                    del cached[cached_arg]
        
        waiter = self._waiters.pop(arg)
        if (waiter is not None):
            waiter.set_result(result)
        
        return result
    
    def __repr__(self):
        """Returns the cacher's representation."""
        repr_parts = [
            self.__class__.__name__,
            '(func=',
            repr(self.func),
            ', timeout=',
            repr(self.timeout),
        ]
        
        rate_limit_group, rate_limit_limiter = self._rate_limit_proxy_args
        
        repr_parts.append(', rate_limit_group=')
        repr_parts.append(repr(rate_limit_group))
        
        if (rate_limit_limiter is not None):
            repr_parts.append(', rate_limit_limiter=')
            repr_parts.append(repr(rate_limit_limiter))
        
        repr_parts.append(')')
        
        return ''.join(repr_parts)
    
    __call__ = execute



class MultiClientMessageDeleteSequenceSharder:
    """
    Helper class of multi client message sequence deleter.
    
    Attributes
    ----------
    client : ``Client``
        The respective client.
    can_read_message_history : `int`
        Whether the `client` can read message history at the respective channel.
    can_manage_messages : `int`
        Whether the respective client can manage messages at the respective channel.
    delete_mass_task : `None` or ``Task``
        Task of bulk deleting messages.
    delete_new_task : `None` or ``Task``
        Task of deleting new or own messages.
    delete_old_task : `None` or ``Task``
        task of deleting other's old messages.
    """
    __slots__ = ('can_manage_messages', 'can_read_message_history', 'client', 'delete_mass_task', 'delete_new_task',
        'delete_old_task', )
    
    def __new__(cls, client, channel):
        """
        Creates a new helper instance of a multi client message sequence deleter.
        
        Parameters
        ----------
        client : ``Client``
            A client who would execute the delete task.
        channel : ``ChannelTextBase``
            Channel, from where the client would delete messages.
        
        Returns
        -------
        self : `None` or ``MultiClientMessageDeleteSequenceSharder``
            If the respective client could not contribute to any task, returns `None`.
        """
        permissions = channel.cached_permissions_for(client)
        if not permissions&PERMISSION_MASK_VIEW_CHANNEL:
            return None
        
        self = object.__new__(cls)
        self.client = client
        self.can_read_message_history = permissions.can_read_message_history
        self.can_manage_messages = permissions.can_manage_messages
        
        self.delete_mass_task = None
        self.delete_new_task = None
        self.delete_old_task = None
        
        return self
    

class WaitForHandler:
    """
    O(n) event waiter. Added as an event handler by ``Client.wait_for``.
    
    Attributes
    ----------
    waiters : `dict` of (``Future``, `callable`) items
        A dictionary which contains the waiter futures and the respective checks.
    """
    __slots__ = ('waiters', )
    
    def __init__(self):
        """
        Creates a new ``WaitForHandler`` instance.
        """
        self.waiters = {}
    
    async def __call__(self, client, *args):
        """
        Runs the checks of the respective event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who received the respective events.
        args : `tuple` of `Any`
            Other received parameters by the event.
        """
        for future, check in self.waiters.items():
            try:
                result = check(*args)
            except BaseException as err:
                future.set_exception_if_pending(err)
            else:
                if isinstance(result, bool):
                    if result:
                        if len(args) == 1:
                            args = args[0]
                    else:
                        return
                else:
                    args = (*args, result)
                
                future.set_result_if_pending(args)


def _check_is_client_duped(client, client_id):
    """
    Checks whether the client is duplicated.
    
    Raises
    ------
    RuntimeError
        Creating the same client multiple times is not allowed.
    """
    try:
        other_client = CLIENTS[client_id]
    except KeyError:
        return
    
    if other_client is not client:
        raise RuntimeError(f'Creating the same client multiple times is not allowed; {client!r} already exists:, '
            f'{other_client!r}.')



def _message_delete_multiple_private_task_message_id_iterator(groups):
    """
    `message_id` iterator used by ``_message_delete_multiple_private_task``.
    
    This function is a generator.
    
    Parameters
    ----------
    groups : `tuple` (`deque` of (`bool`, `int`), `deque` of `int`, `deque` of `int`)
        `deque`-s, which contain message identifiers depending in which rate limit group they are bound to.
    
    Yields
    ------
    message_id : `int`
    """
    message_group_new, message_group_old, message_group_old_own = groups
    for item in message_group_new:
        yield item[1]
    
    yield from message_group_old
    yield from message_group_old_own


async def _message_delete_multiple_private_task(client, channel_id, groups, reason):
    """
    Internal task used by ``Client.message_delete_multiple``.
    
    This method is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    channel_id : `int`
        The channel's identifier, where the messages are.
    groups : `tuple` (`deque` of (`bool`, `int`), `deque` of `int`, `deque` of `int`)
        `deque`-s, which contain message identifiers depending in which rate limit group they are bound to.
    reason : `None` or `str`
        Additional reason which would show up in the guild's audit logs.
    
    Raises
    ------
    ConnectionError
        No internet connection.
    DiscordException
        If any exception was received from the Discord API.
    """
    for message in _message_delete_multiple_private_task_message_id_iterator(groups):
        await client.http.message_delete(channel_id, message.id, reason)


async def _message_delete_multiple_task(client, channel_id, groups, reason):
    """
    Internal task used by ``Client.message_delete_multiple``.
    
    This method is a coroutine.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier, where the messages are.
    groups : `tuple` (`deque` of (`bool`, `int`), `deque` of `int`, `deque` of `int`)
        `deque`-s, which contain message identifiers depending in which rate limit group they are bound to.
    reason : `None` or `str`
        Additional reason which shows up in the respective guild's audit logs.
    
    Raises
    ------
    ConnectionError
        No internet connection.
    DiscordException
        If any exception was received from the Discord API.
    """
    message_group_new, message_group_old, message_group_old_own = groups
    
    tasks = []
    
    delete_mass_task = None
    delete_new_task = None
    delete_old_task = None
    
    while True:
        if delete_mass_task is None:
            message_limit = len(message_group_new)
            
            # 0 is all good, but if it is more, lets check them
            if message_limit:
                message_ids = []
                message_count = 0
                limit = int((time_now()-1209590.)*1000.-DISCORD_EPOCH)<<22 # 2 weeks - 10s
                
                while message_group_new:
                    own, message_id = message_group_new.popleft()
                    if message_id > limit:
                        message_ids.append(message_id)
                        message_count += 1
                        if message_count == 100:
                            break
                        continue
                    
                    if (message_id+20971520000) < limit:
                        continue
                    
                    # If the message is really older than the limit, with ignoring the 10 second, then we move it.
                    if own:
                        group = message_group_old_own
                    else:
                        group = message_group_old
                    
                    group.appendleft(message_id)
                    continue
                
                if message_count:
                    if message_count == 1:
                        if (delete_new_task is None):
                            message_id = message_ids[0]
                            delete_new_task = Task(client.http.message_delete(channel_id, message_id, reason), KOKORO)
                            tasks.append(delete_new_task)
                    else:
                        delete_mass_task = Task(
                            client.http.message_delete_multiple(channel_id, {'messages': message_ids}, reason),
                                KOKORO)
                        
                        tasks.append(delete_mass_task)
        
        if delete_old_task is None:
            if message_group_old:
                message_id = message_group_old.popleft()
                delete_old_task = Task(client.http.message_delete_b2wo(channel_id, message_id, reason), KOKORO)
                tasks.append(delete_old_task)
        
        if delete_new_task is None:
            if message_group_new:
                group = message_group_new
            elif message_group_old_own:
                group = message_group_old_own
            else:
                group = None
            
            if (group is not None):
                message_id = message_group_old_own.popleft()
                delete_new_task = Task(client.http.message_delete(channel_id, message_id, reason), KOKORO)
                tasks.append(delete_new_task)
        
        if not tasks:
            # It can happen, that there are no more tasks left,  at that case we check if there is more message
            # left. Only at `message_group_new` can be anymore message, because there is a time interval of 10
            # seconds, what we do not move between categories.
            if not message_group_new:
                break
            
            # We really have at least 1 message at that interval.
            own, message_id = message_group_new.popleft()
            # We will delete that message with old endpoint if not own, to make
            # Sure it will not block the other endpoint for 2 minutes with any chance.
            if own:
                delete_new_task = Task(client.http.message_delete(channel_id, message_id, reason), KOKORO)
            else:
                delete_old_task = Task(client.http.message_delete_b2wo(channel_id, message_id, reason), KOKORO)
            
            tasks.append(delete_old_task)
        
        done, pending = await WaitTillFirst(tasks, KOKORO)
        
        for task in done:
            tasks.remove(task)
            try:
                result = task.result()
            except (DiscordException, ConnectionError):
                for task in tasks:
                    task.cancel()
                raise
            
            if task is delete_mass_task:
                delete_mass_task = None
                continue
            
            if task is delete_new_task:
                delete_new_task = None
                continue
            
            if task is delete_old_task:
                delete_old_task = None
                continue
             
            # Should not happen
            continue




async def request_channel_thread_channels(client, guild, channel_id, request_function):
    """
    Gets thread channels trough the discord API with the given http client and function.
    
    This method is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    guild : `Guild``
        The source guild of the channel.
    channel_id : `int`
        The respective guild text channel's identifier.
    request_function : ``CoroutineFunctionType``
        The function to request the threads with.
    
    Returns
    -------
    thread_channels : `list` of ``ChannelThread``
    
    Raises
    ------
    ConnectionError
        No internet connection.
    DiscordException
        If any exception was received from the Discord API.
    """
    thread_channels = []
    
    data = None
    
    while True:
        data = await request_function(client.http, channel_id, data)
        thread_channel_datas = data['threads']
        
        for thread_channel_data in thread_channel_datas:
            thread_channel = ChannelThread(thread_channel_data, client, guild.id)
            thread_channels.append(thread_channel)
        
        thread_user_datas = data['members']
        for thread_user_data in thread_user_datas:
            thread_chanel_id = int(thread_user_data['id'])
            try:
                thread_channel = CHANNELS[thread_chanel_id]
            except KeyError:
                continue
    
            user_id = int(thread_user_data['user_id'])
            user = create_partial_user_from_id(user_id)
            
            thread_user_create(thread_channel, user, thread_user_data)
        
        if not data.get('has_more', True):
            break
        
        if thread_channels:
            before = thread_channels[-1].created_at
        else:
            before = datetime.utcnow()
        
        data = {'before': before}
    
    return thread_channels


class ForceUpdateCache:
    """
    Cache for static `force_update` endpoints.
    
    Attributes
    ----------
    synced : `bool`
        Whether the cache was synced already.
    value : `Any`
        The cached value.
    """
    __slots__ = ('synced', 'value')
    
    def __new__(cls):
        """
        Creates a new ``ForceUpdateCache`` instance.
        """
        self = object.__new__(cls)
        self.synced = False
        self.value = None
        return self
    
    def __repr__(self):
        """Returns the cache's representation."""
        repr_parts = ['<', self.__class__.__name__]
        if self.synced:
            repr_parts.append(' value=')
            repr_parts.append(repr(self.value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def set(self, value):
        """
        Sets value to the cache and marks it as synced.
        
        Parameters
        ----------
        value : `Any`
            The cached value.
        """
        self.value = value
        self.synced = True


def channel_move_sort_key(channel_key):
    """
    Channel move sort key sued at `Client.channel_move`.
    
    Parameters
    ----------
    channel_key : `tuple` of (`int`, `int`, `int`, (`None` or `list` of `...`))
    
        Channel sort keys have the following structure:
        - channel order group
        - position
        - channel identifier
        - sub channel key
    
    Returns
    -------
    order_group : `int`
    """
    return channel_key[0]


def role_move_key(role, position):
    """
    Used at ``Client.role_move`` and ``Client.role_reorder`` to create a json serializable change key from the given
    `role` and `position`.
    
    Parameters
    ----------
    role : ``Role``
        Respective role.
    position : `int`
        The role's new position.
    
    Returns
    -------
    change_key : `dict` of (`str`, `int`)
    """
    return {
        'id': role.id,
        'position': position,
    }


def role_reorder_valid_roles_sort_key(item):
    """
    Kes used inside of ``Client.role_reorder`` to sort out roles based on their position.
    
    Parameters
    ----------
    item : `tuple` of (``Role``, `int`)
        
        The item contains the following elements:
        - role
        - position
    """
    return item[1]


def application_command_autocomplete_choice_parser(choices):
    """
    Tries to build application command autocomplete choices from the given `choices`.
    
    Parameters
    ----------
    choices : `None` or `iterable` of (`str`, `int`, `float`)
        Application command autocomplete choices to parse.
    
    Returns
    -------
    choices_processed : `list` of `dict` of (`str`, `str`)
        Json serializable application command autocomplete choices.
    
    Raises
    ------
    TypeError
        - If `choices` is neither `None`, `tuple`, `list`, `set` ,`dict` nor other `iterable`.
        - If a choice is neither `str`, `int`, `float` nor a `name` - `value` pair.
    """
    choices_processed = []
    
    if choices is None:
        pass
    
    elif isinstance(choices, (list, tuple)):
        for choice in choices:
            choices_processed.append(application_command_autocomplete_choice_validator(choice))
    
    elif isinstance(choices, set):
        for choice in choices:
            choices_processed.append(application_command_autocomplete_choice_validator(choice))
        
        choices_processed.sort(key=application_command_autocomplete_choice_sort_key)
    
    elif isinstance(choices, dict):
        for choice_item in choices.items():
            choices_processed.append(application_command_autocomplete_choice_validator_tuple_item(choice_item))
        
        choices_processed.sort(key=application_command_autocomplete_choice_sort_key)
    
    elif hasattr(type(choices), '__iter__'):
        for choice in choices:
            choices_processed.append(application_command_autocomplete_choice_validator(choice))
        
        # Should we sort it?
    
    else:
        raise TypeError(f'`{choices}` can be either `None` or `list`, `tuple`, `set`, `dict` including any other '
            f'iterables, got {choices.__class__.__name__}; {choices!r}.')
    
    del choices_processed[20:]
    
    return choices_processed


def application_command_autocomplete_choice_validator(choice):
    """
    Validates an application autocomplete choice given as a tuple item of 2 elements.
    
    Parameters
    ----------
    choice : `str`, `int`, `float`, `tuple` (`str`, (`str`, `int`, `float`))
        The choice or a `name` - `value` pair.
    
    Returns
    -------
    choice : `dict` of (`str`, `str`)
        Json serializable application command autocomplete choice.
    
    Raises
    ------
    TypeError
        If `choice`'s type is incorrect.
    """
    if isinstance(choice, tuple):
        choice_tuple_length = len(choice)
        if choice_tuple_length == 2:
            return application_command_autocomplete_choice_validator_tuple_item(choice)
        
        raise TypeError(f'Tuple autocomplete choice can have length `2`, got {choice_tuple_length}; {choice!r}.')
    
    if isinstance(choice, str):
        pass
    elif isinstance(choice, (int, float)):
        choice = str(choice)
    else:
        raise TypeError(f'An autocomplete choice can be either a `name` - `value` pair or a `str`, `int` or a `float` '
            f'instance, got {choice.__class__.__name__}; {choice!r}.')
    
    return application_command_autocomplete_choice_builder(choice, choice)


def application_command_autocomplete_choice_validator_tuple_item(item):
    """
    Validates an application autocomplete choice given as a tuple item of 2 elements.
    
    Parameters
    ----------
    item : `tuple` (`str`, (`str`, `int`, `float`))
        A `name` - `value` pair.
    
    Returns
    -------
    choice : `dict` of (`str`, `str`)
        Json serializable application command autocomplete choice.
    
    Raises
    ------
    TypeError
        - If `item[0]` is not `str` instance.
        - If `item[1]` is neither `str`, `int` nor `float` instance.
    """
    name, value = item
    if not isinstance(name, str):
        raise TypeError(f'Tuple item autocomplete choice name should be `str` instance, got '
            f'{name.__class__.__name__}; {name!r}.')
    
    if isinstance(value, str):
        pass
    elif isinstance(value, (int, float)):
        value = str(value)
    else:
        raise TypeError(f'Tuple item autocomplete choice value should be either `str`, `int` or `float` instance, got '
            f'{value.__class__.__name__}, {value!r}.')
    
    return application_command_autocomplete_choice_builder(name, value)


def application_command_autocomplete_choice_builder(name, value):
    """
    Builds an application autocomplete choice from the given `name` - `value` pair.
    
    Parameters
    ----------
    name : `str`
        Choice name.
    value : `str`, `int` or `float`
        Choice value.
    
    Returns
    -------
    choice : `dict` of (`str`, `str`)
        Json serializable application command autocomplete choice.
    """
    return {
        'name': name,
        'value': value,
    }


def application_command_autocomplete_choice_sort_key(choice):
    """
    Choice sort key used to sort unsorted application command autocomplete keys.
    
    Parameters
    ----------
    choice : `dict` of (`str`, `str`)
        Json serializable application command autocomplete choice.
    
    Returns
    -------
    choice_sort_key : `str`
        The choice's name.
    """
    return choice['name']


def try_get_user_id_from_token(token):
    """
    Tries to get user id from the given user token.
    
    Parameters
    ----------
    token : `str`
        The user's token.
    
    Returns
    -------
    user_id : `int`
        Returns `0` if parsing failed.
    """
    dot_index = token.find('.')
    if (dot_index > 0):
        token_base64 = token[:dot_index]
        
        try:
            token_string = b64decode(token_base64)
        except Base64DecodeError:
            user_id = 0
        else:
            try:
                user_id = int(token_string)
            except ValueError:
                user_id = 0
    else:
        user_id = 0
    
    return user_id



__all__ = ('ActivityAssets', 'ActivityBase', 'ActivityParty', 'ActivitySecrets', 'ActivityTimestamps',
    'ACTIVITY_TYPES',)

from datetime import datetime

from ..utils import DISCORD_EPOCH_START, DATETIME_FORMAT_CODE, unix_time_to_datetime, datetime_to_unix_time
from ..color import Color

from . import activity_types as ACTIVITY_TYPES

DEFAULT_CUSTOM_ID = 'UNKNOWN'

CUSTOM_IDS = {
    ACTIVITY_TYPES.spotify: 'spotify:1',
    ACTIVITY_TYPES.custom: 'custom',
}

ACTIVITY_TYPE_NAMES = {
    ACTIVITY_TYPES.game: 'game',
    ACTIVITY_TYPES.stream: 'stream',
    ACTIVITY_TYPES.spotify: 'spotify',
    ACTIVITY_TYPES.watching: 'watching',
    ACTIVITY_TYPES.custom: 'custom',
    ACTIVITY_TYPES.competing: 'competing',
}


class ActivityTimestamps:
    """
    Represents an activity's timestamp field.
    
    Attributes
    ----------
    end : `None` or `datetime`
        When the activity. Defaults to `None`.
    start : `None` or `datetime`
       When the activity starts. Defaults to `None`.
    """
    __slots__ = ('end', 'start',)
    
    def __new__(cls, *, start=None, end=None):
        """
        Creates a new activity timestamp with the given parameters.
        
        Parameters
        ----------
        end : `None` or `datetime`, Optional (Keyword only)
            When the activity. Defaults to `None`.
        start : `None` or `datetime`, Optional (Keyword only)
           When the activity starts. Defaults to `None`.
        
        Raises
        ------
        AssertionError
            - If `start` is nether `None` nor `datetime`.
            - If `end` is neither `None` nor `datetime`.
        """
        if __debug__:
            if (start is not None) and (not isinstance(start, datetime)):
                raise AssertionError(f'`start` can be either `None` or `datetime`, got {start.__class__.__name__}.')

            if (end is not None) and (not isinstance(end, datetime)):
                raise AssertionError(f'`end` can be either `None` or `datetime`, got {end.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.start = start
        self.end = end
        return self
    
    
    @classmethod
    def from_data(cls, timestamps_data):
        """
        Creates a new activity timestamp object from the given data.
        
        Parameters
        ----------
        timestamps_data : `dict` of (`str`, `Any`) items
            Activity timestamp data.
        """
        start = timestamps_data.get('start', None)
        if (start is not None):
            start = unix_time_to_datetime(start)
        
        end = timestamps_data.get('end', None)
        if (end is not None):
            end = unix_time_to_datetime(end)
        
        self = object.__new__(cls)
        self.start = start
        self.end = end
        return self
    
    
    def __repr__(self):
        """Returns the activity timestamp's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        start = self.start
        if (start is not None):
            repr_parts.append(' start=')
            repr_parts.append(start.__format__(DATETIME_FORMAT_CODE))
            field_added = True
        else:
            field_added = False
        
        end = self.end
        if (end is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' end=')
            repr_parts.append(start.__format__(DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two activity timestamps are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.start != other.start:
            return False
        
        if self.end != other.end:
            return False
        
        return True
    
    
    def to_data(self):
        """
        Serializes the activity timestamp.
        
        Returns
        -------
        timestamps_data : `dict` of (`str`, `Any`) items
        """
        timestamps_data = {}
        
        start = self.start
        if (start is not None):
            timestamps_data['start'] = datetime_to_unix_time(start)
        
        end = self.end
        if (end is not None):
            timestamps_data['end'] = datetime_to_unix_time(end)
        
        return timestamps_data


class ActivityAssets:
    """
    Represents a discord activity asset.
    
    Attributes
    ----------
    image_large : `None` or `str`
        The id of the activity's large asset to display. Defaults to `None`.
    image_small : `None` or `str`
        The id of the activity's small asset to display. Defaults to `None`.
    text_large : `None` or `str`
        The hover text of the large asset. Defaults to `None`.
    text_small : `None` or `str`
        The hover text of the small asset. Defaults to `None`.
    """
    __slots__ = ('image_large', 'image_small', 'text_large', 'text_small',)
    
    def __new__(cls, *, image_large=None, image_small=None, text_large=None, text_small=None):
        """
        Creates a new ``ActivityAssets`` instance from the given parameters.
        
        Parameters
        ----------
        image_large : `None` or `str`, Optional (Keyword only)
            The id of the activity's large asset to display. Defaults to `None`.
        image_small : `None` or `str`, Optional (Keyword only)
            The id of the activity's small asset to display. Defaults to `None`.
        text_large : `None` or `str`, Optional (Keyword only)
            The hover text of the large asset. Defaults to `None`.
        text_small : `None` or `str`, Optional (Keyword only)
            The hover text of the small asset. Defaults to `None`.
        
        Raises
        ------
        AssertionError
            - If `image_large` is neither `None` nor `str`.
            - If `image_small` is neither `None` nor `str`.
            - If `text_large` is neither `None` nor `str`.
            - If `text_small` is neither `None` nor `str`.
        """
        if (image_large is not None):
            if __debug__:
                if (not isinstance(image_large, str)):
                    raise AssertionError(f'`image_large` can be either `None` or `str`, got '
                        f'{image_large.__class__.__name__}.')
            
            if (not image_large):
                image_large = None
            
            
        if (image_small is not None):
            if __debug__:
                if (not isinstance(image_small, str)):
                    raise AssertionError(f'`image_small` can be either `None` or `str`, got '
                        f'{image_small.__class__.__name__}.')
            
            if (not image_small):
                image_small = None
            
        
        if (text_large is not None):
            if __debug__:
                if (not isinstance(text_large, str)):
                    raise AssertionError(f'`text_large` can be either `None` or `str`, got '
                        f'{text_large.__class__.__name__}.')
            
            if (not text_large):
                text_large = None
        
        if (text_small is not None):
            if __debug__:
                if (not isinstance(text_small, str)):
                    raise AssertionError(f'`image_large` can be either `None` or `str`, got '
                        f'{text_small.__class__.__name__}.')
            
            if (not text_small):
                text_small = None
        
        self = object.__new__(cls)
        self.image_large = image_large
        self.image_small = image_small
        self.text_large = text_large
        self.text_small = text_small
        return self
    
    @classmethod
    def from_data(cls, assets_data):
        """
        Creates a new activity asset object from the given data.
        
        Parameters
        ----------
        assets_data : `dict` of (`str`, `Any`) items
            Activity asset data.
        """
        self = object.__new__(cls)
        self.image_large = assets_data.get('large_image', None)
        self.image_small = assets_data.get('small_image', None)
        self.text_large = assets_data.get('large_text', None)
        self.text_small = assets_data.get('small_text', None)
        return self
    
    def __repr__(self):
        """Returns the activity asset's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        image_large = self.image_large
        if (image_large is not None):
            repr_parts.append(' image_large=')
            repr_parts.append(repr(image_large))
            field_added = True
        else:
            field_added = False
        
        image_small = self.image_small
        if (image_small is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' image_small=')
            repr_parts.append(repr(image_small))
        
        text_large = self.text_large
        if (text_large is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' text_large=')
            repr_parts.append(repr(text_large))
        
        text_small = self.text_small
        if (text_small is not None):
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' text_small=')
            repr_parts.append(repr(text_small))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two activity assets are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.image_large != other.image_large:
            return False
        
        if self.image_small != other.image_small:
            return False
        
        if self.text_large != other.text_large:
            return False
        
        if self.text_small != other.text_small:
            return False
        
        return True
    
    
    def to_data(self):
        """
        Serializes the activity asset.
        
        Returns
        -------
        timestamp_data : `dict` of (`str`, `Any`) items
        """
        assets_data = {}
        
        image_large = self.image_large
        if (image_large is not None):
            assets_data['large_image'] = image_large
        
        image_small = self.image_small
        if (image_small is not None):
            assets_data['small_image'] = image_small
        
        text_large = self.text_large
        if (text_large is not None):
            assets_data['large_text'] = text_large
        
        text_small = self.text_small
        if (text_small is not None):
            assets_data['small_text'] = text_small
        
        return assets_data


class ActivityParty:
    """
    Represents a discord activity party.
    
    Attributes
    ----------
    id : `None` or `str`
        The party's id, which in the player is. Defaults to `None`.
    size : `int`
        The party's maximal size, which in the player is. Defaults to `0`.
    max : `int`
        The party's actual size, which in the player is. Defaults to `0`.
    """
    __slots__ = ('id', 'size', 'max',)
    
    def __new__(cls, *, id_=None, size=0, max_=0):
        """
        Creates a new activity party instance form the given parameters.
        
        Parameters
        ----------
        id_ : `None` or `str`, Optional (Keyword only)
            The party's id, which in the player is. Defaults to `None`.
        size : `int`, Optional (Keyword only)
            The party's maximal size, which in the player is. Defaults to `0`.
        max_ : `int`, Optional (Keyword only)
            The party's actual size, which in the player is. Defaults to `0`.
        
        Raises
        ------
        AssertionError
            - If `id_` is neither `None` nor `str` instance.
            - if `size` is not `int` instance.
            - If `max_` is not `int` instance.
            - If `size` is negative`.
            - If `max_` is negative.
        """
        if (id_ is not None):
            if __debug__:
                if (not isinstance(id_, int)):
                    raise AssertionError(f'`id_` can be either `None` or `str`, got {id_.__class__.__name__}.')
            
            if (not id_):
                id_ = None
        
        if __debug__:
            if (not isinstance(size, int)):
                raise AssertionError(f'`size` can be `int` instance, got {size.__class__.__name__}.')
            
            if (not isinstance(max_, int)):
                raise AssertionError(f'`max_` can be `int` instance, got {max_.__class__.__name__}.')
            
            if (size < 0):
                raise AssertionError(f'`size` cannot be negative, got {size!r}.')
            
            if (max_ < 0):
                raise AssertionError(f'`max_` cannot be negative, got {max_!r}.')
        
        self = object.__new__(cls)
        self.id = id_
        self.size = size
        self.max = max_
        return self
    
    
    @classmethod
    def from_data(cls, party_data):
        """
        Creates a new activity party object from the given data.
        
        Parameters
        ----------
        party_data : `dict` of (`str`, `Any`) items
            Activity party data.
        """
        
        self = object.__new__(cls)
        self.id = party_data.get('id', None)
        
        try:
            size, max_ = party_data['size']
        except KeyError:
            size = 0
            max_ = 0
        
        self.size = size
        self.max = max_
        return self
    
    def __repr__(self):
        """Returns the activity party's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        id_ = self.id
        if (id_ is not None):
            repr_parts.append(' id_=')
            repr_parts.append(repr(id_))
            field_added = True
        else:
            field_added = False
        
        size = self.size
        max_ = self.max
        if size or max_:
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' size=')
            repr_parts.append(repr(size))
            repr_parts.append(', max=')
            repr_parts.append(repr(max_))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def __eq__(self, other):
        """Returns whether the two activity parties are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.id != other.id:
            return False
        
        if self.size != other.size:
            return False
        
        if self.max != other.max:
            return False
        
        return True
    
    def to_data(self):
        """
        Serializes the activity party.
        
        Returns
        -------
        timestamp_data : `dict` of (`str`, `Any`) items
        """
        party_data = {}
        
        id_ = self.id
        if (id_ is not None):
            party_data['id'] = id_
        
        size = self.size
        max_ = self.max
        if size or max_:
            party_data['size'] = [size, max_]
        
        return party_data


class ActivitySecrets:
    """
    Represents and activity secret.
    
    Attributes
    ----------
    join : `None` or `str`
        Unique hash given for the match context. Defaults to `None`.
    match : `None` or `str`
        Unique hash for spectate button. Defaults to `None`.
    spectate : `None` or `str`
        Unique hash for chat invites and ask to join. Defaults to `None`.
    """
    __slots__ = ('join', 'match', 'spectate', )
    
    def __new__(cls, *, join=None, match=None, spectate=None):
        """
        Creates a new activity secret from the given parameters.
        
        Parameters
        ----------
        join : `None` or `str`, Optional (Keyword only)
            Unique hash given for the match context. Defaults to `None`.
        match : `None` or `str`, Optional (Keyword only)
            Unique hash for spectate button. Defaults to `None`.
        spectate : `None` or `str`, Optional (Keyword only)
            Unique hash for chat invites and ask to join. Defaults to `None`.
        """
        if (join is not None):
            if __debug__:
                if (not isinstance(join, str)):
                    raise AssertionError(f'`join` can be either `None` or `str`, got {join.__class__.__name__}.')
            
            if (not join):
                join = None
            
            
        if (match is not None):
            if __debug__:
                if (not isinstance(match, str)):
                    raise AssertionError(f'`match` can be either `None` or `str`, got {match.__class__.__name__}.')
            
            if (not match):
                match = None
            
        
        if (spectate is not None):
            if __debug__:
                if (not isinstance(spectate, str)):
                    raise AssertionError(f'`spectate` can be either `None` or `str`, got {spectate.__class__.__name__}.')
            
            if (not spectate):
                spectate = None
        
        self = object.__new__(cls)
        self.join = join
        self.match = match
        self.spectate = spectate
        return self
    
    @classmethod
    def from_data(cls, secrets_data):
        """
        Creates a new activity secret object from the given data.
        
        Parameters
        ----------
        secrets_data : `dict` of (`str`, `Any`) items
            Activity secret data.
        """
        self = object.__new__(cls)
        self.join = secrets_data.get('join', None)
        self.spectate = secrets_data.get('spectate', None)
        self.match = secrets_data.get('match', None)
        return self
    
    def __repr__(self):
        """Returns the activity secret's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        join = self.join
        if (join is not None):
            repr_parts.append(' join=')
            repr_parts.append(repr(join))
            field_added = True
        else:
            field_added = False
        
        spectate = self.spectate
        if (spectate is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' spectate=')
            repr_parts.append(repr(spectate))
        
        match = self.match
        if (match is not None):
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' match=')
            repr_parts.append(repr(match))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def __eq__(self, other):
        """Returns whether the two activity secrets are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.join != other.join:
            return False
        
        if self.spectate != other.spectate:
            return False
        
        if self.match != other.match:
            return False
        
        return True
    
    def to_data(self):
        """
        Serializes the activity secret.
        
        Returns
        -------
        timestamp_data : `dict` of (`str`, `Any`) items
        """
        secrets_data = {}
        
        join = self.join
        if (join is not None):
            secrets_data['join'] = join
        
        spectate = self.spectate
        if (spectate is not None):
            secrets_data['spectate'] = spectate
        
        match = self.match
        if (match is not None):
            secrets_data['match'] = match
        
        return secrets_data


class ActivityBase:
    """
    Base class for activities.
    
    Class Attributes
    ----------------
    created_at : `datetime`
        When the activity was created. Defaults to Discord epoch.
    name : `str` = `'Unknown'`
        The activity's name. Subclasses might overwrite it as member descriptor.
    id : `int` = `0`
        The activity's id. Subclasses might overwrite it as member descriptor.
    type : `int` = `127`
        The activity's type value.
    """
    name = 'Unknown'
    id = 0
    type = 127
    created_at = DISCORD_EPOCH_START
    
    __slots__ = ()
    def __new__(cls, data):
        """
        Creates a new activity. Neither ``ActivityBase`` or it's subclass: ``ActivityUnknown`` cannot be instanced and
        raises `RuntimeError`.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Activity data received from Discord.
        
        Raises
        ------
        RuntimeError
        """
        raise RuntimeError(f'{cls.__name__} cannot be instanced.')
    
    
    def __repr__(self):
        """Returns the activity's representation."""
        return f'<{self.__class__.__name__} name={self.name!r}>'
    
    
    def __hash__(self):
        """Returns the activity's hash value."""
        return self.id
    
    
    def __eq__(self, other):
        """
        Returns whether the two activities are equal.
        
        Subclasses should overwrite it.
        """
        return NotImplemented
    
    @property
    def color(self):
        """
        Returns the activity's color.
        
        Subclasses should overwrite it.
        
        Returns
        -------
        color : ``Color``
        """
        return Color()
    
    
    @property
    def discord_side_id(self):
        """
        Returns the activity's Discord side id. If the activity implements id returns that, else returns it's
        `CUSTOM_ID` class attribute.
        
        Returns
        -------
        activity_id : `str`
        """
        id_ = self.id
        if id_:
            activity_id = self.id.__format__('x')
        else:
            activity_id = CUSTOM_IDS.get(self.type, DEFAULT_CUSTOM_ID)
        
        return activity_id
    
    @property
    def created_at(self):
        """
        When the activity was created. If the creation time was not included, then returns the discord's epoch's start.
        
        Returns
        -------
        created_at : `datetime`
        """
        return DISCORD_EPOCH_START
    
    @classmethod
    def from_data(cls, activity_data):
        """
        Creates a new activity instance from the given activity data.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Received activity data.
        
        Returns
        -------
        activity : `None`
        """
        return None
    
    def _difference_update_attributes(self, data):
        """
        Updates the activity and returns the changes in a `dict` of (`attribute-name`, `old-value`) items.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict`
            Always empty.
        """
        return {}
    
    def _update_attributes(self, data):
        """
        Updates the activity by overwriting it's old attributes.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        pass
    
    def bot_dict(self):
        """
        Converts the activity to json serializable dictionary, which can be sent with bot account to change activity.
        
        Subclasses, which implement it, should overwrite.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    def user_dict(self):
        """
        Converts the activity to json serializable dictionary, which can (?) be sent with user account to change
        activity.
        
        Subclasses, which implement it, should overwrite.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    def full_dict(self):
        """
        Converts the whole activity to a dictionary.
        
        Subclasses, which implement it, should overwrite.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        return {}



__all__ = ('ActivityCustom',)

from ...backend.export import include
from ..utils import DISCORD_EPOCH_START, unix_time_to_datetime, datetime_to_unix_time

from .activity_base import ActivityBase
from . import activity_types as ACTIVITY_TYPES

create_partial_emoji_from_data = include('create_partial_emoji_from_data')

class ActivityCustom(ActivityBase):
    """
    Represents a Discord custom activity.
    
    Attributes
    ----------
    created : `int`
        When the status was created as Unix time in milliseconds. Defaults to `0`.
    emoji : `None` or ``Emoji``
        The emoji of the activity. If it has no emoji, then set as `None`.
    state : `None` or `str`
        The activity's text under it's emoji. Defaults to `None`.
    
    Class Attributes
    ----------------
    id : `int` = `0`
        The activity's id.
    type : `int` = `4`
        The activity's type value.
    """
    __slots__ = ('created_at', 'emoji', 'state', )
    
    type = ACTIVITY_TYPES.custom
    
    @classmethod
    def from_data(cls, activity_data):
        """
        Creates a new ``ActivityCustom`` instance from the given activity data.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Received activity data.
        
        Returns
        -------
        self : ``ActivityCustom``
        """
        self = object.__new__(cls)
        self._update_attributes(activity_data)
        return self
    
    def __hash__(self):
        """Returns the activity's hash value."""
        state = self.state
        emoji = self.emoji
        if (state is None):
            if (emoji is None):
                hash_ = 0
            else:
                hash_ = emoji.id
        else:
            hash_ = hash(state)
            if (emoji is not None):
                hash_ ^=emoji.id
        
        return hash_
    
    @property
    def name(self):
        """
        Returns the activity's display text.
        
        Returns
        -------
        name : `str`
        """
        state = self.state
        emoji = self.emoji
        if (state is None):
            if (emoji is None):
                name = ''
            else:
                name = emoji.as_emoji
        else:
            if (emoji is None):
                name = state
            else:
                name = f'{emoji.as_emoji} {state}'
        
        return name
    
    def _update_attributes(self, activity_data):
        """
        Updates the activity by overwriting it's old attributes.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        self.state = activity_data.get('state', None)
        
        emoji_data = activity_data.get('emoji', None)
        if emoji_data is None:
            emoji = None
        else:
            emoji = create_partial_emoji_from_data(emoji_data)
        self.emoji = emoji
        
        created_at = activity_data.get('created_at', None)
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = unix_time_to_datetime(created_at)
        self.created_at = created_at
    
    def _difference_update_attributes(self, activity_data):
        """
        Updates the activity and returns the changes in a `dict` of (`attribute-name`, `old-value`) items.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        
        Returns
        -------
        changes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
            The returned items might be the following:
            
            +---------------+-----------------------+
            | key           | value                 |
            +===============+=======================+
            | created_at    | `datetime`            |
            +---------------+-----------------------+
            | emoji         | `None` or ``Emoji``   |
            +---------------+-----------------------+
            | state         | `None` or `str`       |
            +---------------+-----------------------+
        """
        old_attributes = {}
        
        state = activity_data.get('state', None)
        if self.state != state:
            old_attributes['state'] = self.state
            self.state = state
        
        emoji_data = activity_data.get('emoji', None)
        if emoji_data is None:
            emoji = None
        else:
            emoji = create_partial_emoji_from_data(emoji_data)
        
        if self.emoji != emoji:
            old_attributes['emoji'] = self.emoji
            self.emoji = emoji
        
        created_at = activity_data.get('created_at', None)
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = unix_time_to_datetime(created_at)
        if self.created_at != created_at:
            old_attributes['created_at'] = self.created_at
            self.created_at = created_at
        
        return old_attributes
    
    def full_dict(self):
        """
        Converts the whole activity to a dictionary.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        activity_data = {
            'name': 'Custom Status',
            'id': 'custom',
        }
        
        emoji = self.emoji
        if (emoji is not None):
            emoji_data = {}
            if emoji.is_custom_emoji():
                emoji_data['name'] = emoji.name
                emoji_data['id'] = emoji.id
                if emoji.animated:
                    emoji_data['animated'] = True
            else:
                emoji_data['name'] = emoji.unicode
            
            activity_data['emoji']=emoji_data
        
        state = self.state
        if (state is not None):
            activity_data['state'] = state
        
        created_at = self.created_at
        if created_at != DISCORD_EPOCH_START:
            activity_data['created_at'] = datetime_to_unix_time(created_at)
        
        return activity_data



import sys
from math import log10, floor
from .. import __package__ as PACKAGE_NAME
from ..__main__ import COMMAND_MAP, COMMAND_NAMES
from ..backend.utils import get_short_executable

PACKAGE = __import__(PACKAGE_NAME)

NAME = 'help'
USAGE = 'h | help *command*'

HELP = (
    f'Either lists the available command, or shows the command\'s usage.\n'
)

def show_command(command_name):
    command_file_name = COMMAND_MAP.get(command_name.lower(), None)
    
    output_parts = []
    if command_file_name is None:
        output_parts = []
        
        output_parts.append('No command is added for: ')
        output_parts.append(repr(command_name))
        output_parts.append('\n')
        output_parts.append('Try using "$ ')
        output_parts.append(get_short_executable())
        output_parts.append(' ')
        output_parts.append(PACKAGE_NAME)
        output_parts.append(' help" to list all available commands\n.')
    else:
        __import__(f'{PACKAGE_NAME}.main.{command_file_name}')
        command_module = getattr(PACKAGE.main, command_file_name)
        name = command_module.NAME
        usage = command_module.USAGE
        help_ = command_module.HELP
        
        output_parts.append('Help for: "')
        output_parts.append(name)
        output_parts.append('"\n\nUsage: "$ ')
        output_parts.append(get_short_executable())
        output_parts.append(' ')
        output_parts.append(PACKAGE_NAME)
        output_parts.append(' ')
        output_parts.append(usage)
        output_parts.append('"\n\n')
        output_parts.append(help_)
    
    output = ''.join(output_parts)
    sys.stderr.write(output)


def list_commands():
    output_parts = []
    output_parts.append('Available commands:\n\n')
    
    command_count = len(COMMAND_NAMES)
    index_adjust = floor(log10(command_count))+1
    index = 0
    while index != command_count:
        command_name = COMMAND_NAMES[index]
        index += 1
        output_parts.append(str(index).rjust(index_adjust))
        output_parts.append('.: ')
        output_parts.append(command_name)
        output_parts.append('\n')
    
    output_parts.append('\n')
    output_parts.append('Use "$ ')
    output_parts.append(get_short_executable())
    output_parts.append(' ')
    output_parts.append(PACKAGE_NAME)
    output_parts.append(' help ')
    output_parts.append('*command*')
    output_parts.append('" for more information.\n')
    
    output = ''.join(output_parts)
    sys.stderr.write(output)


def __main__():
    system_parameters = sys.argv
    if len(system_parameters) < 3:
        command_parameter = None
    else:
        command_parameter = system_parameters[2].lower()
    
    if command_parameter is None:
        list_commands()
    else:
        show_command(command_parameter)



from .integration import *
from .integration_account import *
from .integration_application import *
from .integration_detail import *
from .preinstanced import *
from .utils import *

__all__ = (
    *integration.__all__,
    *integration_account.__all__,
    *integration_application.__all__,
    *integration_detail.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)



__all__ = ('PermissionOverwriteTargetType',)

from ...env import API_VERSION
from ...backend.utils import set_docs
from ..bases import PreinstancedBase, Preinstance as P


class PermissionOverwriteTargetType(PreinstancedBase):
    INSTANCES = {}
    VALUE_TYPE = str if API_VERSION in (6, 7) else int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    role = P('role' if API_VERSION in (6, 7) else 0, 'role')
    user = P('member' if API_VERSION in (6, 7) else 1, 'user')


set_docs(PermissionOverwriteTargetType,
    """
    Represents a permission overwrite's target's type.
    
    Attributes
    ----------
    name : `str`
        The name of the permission overwrite target type.
    value : `int`
        The identifier value the permission overwrite target type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``PermissionOverwriteTargetType``) items
        Stores the predefined ``PermissionOverwriteTargetType`` instances. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `str`
        The permission overwrite target types' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the permission overwrite target types.
    
    Every predefined permission overwrite target type can be accessed as class attribute as well:
    
    +-----------------------+---------------+---------------+
    | Class attribute name  | name          | value         |
    +=======================+===============+===============+
    | role                  | role          | `'role'`      |
    +-----------------------+---------------+---------------+
    | user                  | user          | `'member'`    |
    +-----------------------+---------------+---------------+
    """
    if API_VERSION in (6, 7) else
    """
    Represents a permission overwrite's target's type.
    
    Attributes
    ----------
    name : `str`
        The name of the permission overwrite target type.
    value : `int`
        The identifier value the permission overwrite target type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``PermissionOverwriteTargetType``) items
        Stores the predefined ``PermissionOverwriteTargetType`` instances. These can be accessed with their `value` as
        key.
    VALUE_TYPE : `type` = `int`
        The permission overwrite target types' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the permission overwrite target types.
    
    Every predefined permission overwrite target type can be accessed as class attribute as well:
    
    +-----------------------+---------------+---------------+
    | Class attribute name  | name          | value         |
    +=======================+===============+===============+
    | role                  | role          | 0             |
    +-----------------------+---------------+---------------+
    | user                  | user          | 1             |
    +-----------------------+---------------+---------------+
    """
)
